"""Errors raised by the library"""


class Error(Exception):
    pass


class RemoteTraceback(Error):
    """An instance of this class is available on the
     __context__ attribute of the last remote exception in an
     exception chain that occurred in a ProcessPool.
     """
    def __init__(self, traceback_lines):
        self._traceback_lines = traceback_lines

    @property
    def traceback_lines(self):
        return self._traceback_lines

    def __str__(self):
        tbs = "".join(self._traceback_lines)
        return '\n"""\n{}"""'.format(tbs)


class InvalidStateError(Error):
    pass


class CancelledError(Error):
    pass


class BrokenPoolError(Error):
    pass


class InitializerError(BrokenPoolError):
    pass


class FinalizerError(BrokenPoolError):
    pass



"""The abstract base class for all pools is defined here,
as well as the GlobalShutdown class."""
import time
import atexit
import threading
import itertools
import multiprocessing as mp
from abc import ABC, abstractmethod
from collections import deque
from asyncpal import errors
from asyncpal import misc
from asyncpal.future import Future, Status, as_done, FutureFilter
from asyncpal.worker import WorkerType


__all__ = ["Pool", "IDLE_TIMEOUT", "MP_CONTEXT",
           "WINDOWS_MAX_PROCESS_WORKERS"]


IDLE_TIMEOUT = 60
MP_CONTEXT = mp.get_context("spawn")
WINDOWS_MAX_PROCESS_WORKERS = 60


class Pool(ABC):
    def __init__(self, worker_type, max_workers, name,
                 idle_timeout, initializer, init_args,
                 init_kwargs, finalizer, final_args,
                 final_kwargs, max_tasks_per_worker,
                 task_queue, mp_task_queue=None,
                 message_queue=None, mp_context=None):
        # constructor args and kwargs
        self._worker_type = WorkerType(worker_type)
        self._max_workers = max_workers
        self._name = name
        self._idle_timeout = idle_timeout
        self._initializer = initializer
        init_args_kwargs = misc.ensure_args_kwargs(init_args, init_kwargs)
        self._init_args, self._init_kwargs = init_args_kwargs
        final_args_kwargs = misc.ensure_args_kwargs(final_args, final_kwargs)
        self._final_args, self._final_kwargs = final_args_kwargs
        self._finalizer = finalizer
        self._max_tasks_per_worker = max_tasks_per_worker
        self._mp_context = mp_context
        # task queue
        self._task_queue = task_queue
        self._mp_task_queue = mp_task_queue
        # task filter thread
        self._filter_thread = None
        # message queue and message thread for Process based work
        self._message_queue = message_queue
        self._message_thread = None
        # mutexes
        self._vars_lock = self._create_lock()
        self._futures_lock = self._create_lock()
        self._workers_lock = self._create_lock()
        self._pool_lock = self._create_lock()
        #
        self._workers = dict()
        self._inactive_workers = list()
        self._stored_futures = dict()
        self._is_broken = False
        self._cancelled_tasks = list()
        self._is_closed = threading.Event()
        self._is_terminated = False
        self._stored_exception = None
        self._monotonic_worker_count = 0
        self._monotonic_task_count = 0
        self._setup()

    @property
    def worker_type(self):
        return self._worker_type

    @property
    def max_workers(self):
        with self._vars_lock:
            return self._max_workers

    @property
    def name(self):
        return self._name

    @property
    def idle_timeout(self):
        with self._vars_lock:
            return self._idle_timeout

    @idle_timeout.setter
    def idle_timeout(self, val):
        with self._vars_lock:
            self._idle_timeout = val

    @property
    def initializer(self):
        with self._vars_lock:
            return self._initializer

    @property
    def init_args(self):
        with self._vars_lock:
            return self._init_args

    @property
    def init_kwargs(self):
        with self._vars_lock:
            return self._init_kwargs

    @property
    def finalizer(self):
        with self._vars_lock:
            return self._finalizer

    @property
    def final_args(self):
        with self._vars_lock:
            return self._final_args

    @property
    def final_kwargs(self):
        with self._vars_lock:
            return self._final_kwargs

    @property
    def max_tasks_per_worker(self):
        return self._max_tasks_per_worker

    @property
    def mp_context(self):
        return self._mp_context

    @property
    def workers(self):
        with self._vars_lock:
            return tuple(self._workers.values())

    @property
    def cancelled_tasks(self):
        with self._vars_lock:
            return tuple(self._cancelled_tasks)

    @property
    def is_closed(self):
        return self._is_closed.is_set()

    @property
    def is_terminated(self):
        with self._vars_lock:
            return self._is_terminated

    @property
    def is_broken(self):
        with self._vars_lock:
            return self._is_broken

    def check(self):
        """
        Check the pool

        [except]
        - RuntimeError: raised if the pool is closed
        - BrokenPoolError: raised if the pool is broken
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()

    def run(self, target, /, *args, **kwargs):
        """
        Submit the task to the pool, and return
        the result (or re-raise the exception raised by the callable)

        [param]
        - target: callable
        - args: args to pass to the callable
        - kwargs: kwargs to pass to the callable

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: exception that might be raised by the task itself
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            return self._submit_task(target, *args, **kwargs).collect()

    def submit(self, target, /, *args, **kwargs):
        """
        Submit the task to the pool, and return
        a future object

        [param]
        - target: callable
        - args: args to pass to the callable
        - kwargs: kwargs to pass to the callable

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: exception that might be raised by the task itself
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            return self._submit_task(target, *args, **kwargs)

    def map(self, target, *iterables, chunk_size=1, buffer_size=1,
            keep_order=True, timeout=None):
        """
        Perform a Map operation lazily and return an iterator
        that iterates over the results.
        Beware, a remote exception will be reraised here

        [param]
        - target: callable
        - iterables: iterables to pass to the target
        - chunk_size: max length for a chunk
        - buffer_size: the buffer_size. A bigger size will consume more memory
            but the overall operation will be faster
        - keep_order: whether the original order should be kept or not
        - timeout: None or a timeout (int or float) value in seconds

        [return]
        Returns an iterator

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: any remote exception
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            if chunk_size == 1 and keep_order:
                return self._map_lazy(target, zip(*iterables),
                                      buffer_size=buffer_size,
                                      timeout=timeout)
            elif chunk_size == 1:
                return self._map_lazy_unordered(target, zip(*iterables),
                                                buffer_size=buffer_size,
                                                timeout=timeout)
            elif keep_order:
                return self._map_lazy_chunked(target, zip(*iterables),
                                              chunk_size=chunk_size,
                                              buffer_size=buffer_size,
                                              timeout=timeout)
            else:
                return self._map_lazy_chunked_unordered(target, zip(*iterables),
                                                        chunk_size=chunk_size,
                                                        buffer_size=buffer_size,
                                                        timeout=timeout)

    def map_unordered(self, target, *iterables,
                      chunk_size=1, buffer_size=1,
                      timeout=None):
        """Same as map with 'keep_order' set to False"""
        return self.map(target, *iterables, chunk_size=chunk_size,
                        buffer_size=buffer_size, keep_order=False,
                        timeout=timeout)

    def map_all(self, target, *iterables, chunk_size=1,
                keep_order=True, timeout=None):
        """
        Perform a Map operation eagerly and return an iterator
        that iterates over the results.
        Using this method instead of the `map` method might cause high memory usage.
        Beware, a remote exception will be reraised here

        [param]
        - target: callable
        - iterables: iterables to pass to the target
        - chunk_size: max length for a chunk
        - keep_order: whether the original order should be kept or not
        - timeout: None or a timeout (int or float) value in seconds

        [return]
        Returns an iterator that iterates over the results.

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: any remote exception
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            if chunk_size == 1:
                return self._map_eager(target, zip(*iterables),
                                       keep_order=keep_order,
                                       timeout=timeout)
            else:
                return self._map_eager_chunked(target, zip(*iterables),
                                               chunk_size=chunk_size,
                                               keep_order=keep_order,
                                               timeout=timeout)

    def map_all_unordered(self, target, *iterables, chunk_size=1,
                         timeout=None):
        """Same as map with 'keep_order' set to False"""
        return self.map_all(target, *iterables, chunk_size=chunk_size,
                            keep_order=False, timeout=timeout)

    def starmap(self, target, iterable, chunk_size=1, buffer_size=1,
                keep_order=True, timeout=None):
        """
        Perform a Starmap operation lazily and return an iterator
        that iterates over the results.
        Beware, a remote exception will be reraised here

        [param]
        - target: callable
        - iterable: sequence of args to pass to the target
        - chunk_size: max length for a chunk
        - buffer_size: the buffer_size. A bigger size will consume more memory
            but the overall operation will be faster
        - keep_order: whether the original order should be kept or not
        - timeout: None or a timeout (int or float) value in seconds

        [return]
        Returns an iterator that iterates over the results.

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: any remote exception
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            if chunk_size == 1 and keep_order:
                return self._map_lazy(target, iterable,
                                      buffer_size=buffer_size,
                                      timeout=timeout)
            elif chunk_size == 1:
                return self._map_lazy_unordered(target, iterable,
                                                buffer_size=buffer_size,
                                                timeout=timeout)
            elif keep_order:
                return self._map_lazy_chunked(target, iterable,
                                              chunk_size=chunk_size,
                                              buffer_size=buffer_size,
                                              timeout=timeout)
            else:
                return self._map_lazy_chunked_unordered(target, iterable,
                                                        chunk_size=chunk_size,
                                                        buffer_size=buffer_size,
                                                        timeout=timeout)

    def starmap_unordered(self, target, iterable, chunk_size=1,
                          buffer_size=1, timeout=None):
        """Same as starmap with 'keep_order' set to False"""
        return self.starmap(target, iterable, chunk_size=chunk_size,
                            buffer_size=buffer_size, keep_order=False,
                            timeout=timeout)

    def starmap_all(self, target, iterable, chunk_size=1,
                    keep_order=True, timeout=None):
        """
        Perform a Starmap operation eagerly and return an iterator
        that iterates over the results.
        Using this method instead of the `map` method might cause high memory usage.
        Beware, a remote exception will be reraised here

        [param]
        - target: callable
        - iterable: sequence of args to pass to the target
        - chunk_size: max length for a chunk
        - keep_order: whether the original order should be kept or not
        - timeout: None or a timeout (int or float) value in seconds

        [return]
        Returns an iterator that iterates over the results.

        [except]
        - RuntimeError: raised when the pool is closed
        - BrokenPoolError: raised when the pool is broken
        - Exception: any remote exception
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            if chunk_size == 1:
                return self._map_eager(target, iterable,
                                       keep_order=keep_order,
                                       timeout=timeout)
            else:
                return self._map_eager_chunked(target, iterable,
                                               chunk_size=chunk_size,
                                               keep_order=keep_order,
                                               timeout=timeout)

    def starmap_all_unordered(self, target, iterable, chunk_size=1,
                              timeout=None):
        """Same as starmap_all with 'keep_order' set to False"""
        return self.starmap_all(target, iterable, chunk_size=chunk_size,
                                keep_order=False, timeout=timeout)

    def test(self):
        """Test the pool by creating another pool with the same config
        and doing some computation on it to ensure that it won't break.
        This method might raise a BrokenPoolError exception
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        kwargs = {"max_workers": 1,
                  "idle_timeout": self._idle_timeout,
                  "initializer": self._initializer,
                  "init_args": self._init_args,
                  "init_kwargs": self._init_kwargs,
                  "finalizer": self._finalizer,
                  "final_args": self._final_args,
                  "final_kwargs": self._final_kwargs,
                  "max_tasks_per_worker": self._max_tasks_per_worker}
        pool_class = self.__class__
        if self._worker_type == WorkerType.THREAD:
            name = "test_threadpool_" + str(int(time.time()))
        else:
            name = "test_processpool_" + str(int(time.time()))
            kwargs["mp_context"] = self._mp_context
        kwargs["name"] = name
        with pool_class(**kwargs) as pool:
            it = range(10)
            # with .submit
            r = pool.submit(misc.add, 1, 2).collect()
            assert r == 3
            # with .map
            r = tuple(pool.map(misc.add, it, it, chunk_size=2))
            assert r == tuple(map(misc.add, it, it))
            # with .starmap
            r = tuple(pool.starmap(misc.add, zip(it, it), chunk_size=2))
            assert r == tuple(itertools.starmap(misc.add, zip(it, it)))
            # check !
            pool.check()

    def spawn_workers(self, n=None):
        """Spawn a specific number of workers or the right number
        of workers that is needed

        [param]
        - n: None or an integer.

        [return]
        Returns the number of spawned workers
        """
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            return self._spawn_workers(n)

    def spawn_max_workers(self):
        """Spawn the maximum number of workers"""
        if self._is_closed.is_set():
            raise RuntimeError
        self._ensure_pool_integrity()
        with self._pool_lock:
            self._spawn_workers(self._max_workers)

    def count_workers(self):
        """Returns the number of workers that are alive"""
        if self._is_closed.is_set():
            return 0
        with self._pool_lock:
            return self._count_workers()

    def count_busy_workers(self):
        """Returns the number of busy workers"""
        if self._is_closed.is_set():
            return 0
        with self._pool_lock:
            return self._count_busy_workers()

    def count_free_workers(self):
        """Returns the number of free workers"""
        if self._is_closed.is_set():
            return 0
        with self._pool_lock:
            return self._count_free_workers()

    def count_pending_tasks(self):
        """Returns the number of pending tasks"""
        if self._is_closed.is_set():
            return 0
        with self._pool_lock:
            return self._count_pending_tasks()

    def join(self, timeout=None):
        """
        Join the workers, i.e., wait for workers to end their works, then close them

        [param]
        - timeout: None or a number representing seconds.

        [except]
        - RuntimeError: raised when timeout expires
        """
        if self._is_closed.is_set():
            return False
        self._is_closed.set()
        with self._pool_lock:
            self._notify_workers_to_shutdown()
            r = self._join_workers(timeout)
        self._is_closed.clear()
        return r

    def shutdown(self):
        """
        Close the pool by joining workers and cancelling pending tasks.
        Note that cancelled tasks can be retrieved via the cancelled_tasks property.

        [return]
        Returns False if the pool has already been closed, else returns True.
        """
        if self._is_closed.is_set():
            return False
        with self._pool_lock:
            self._is_closed.set()
            self._cancel_tasks()
            self._notify_workers_to_shutdown()
            self._join_workers()
            self._cleanup_task_queue()
            if self._worker_type == WorkerType.PROCESS:
                self._shutdown_filter_thread()
                self._shutdown_message_thread()
            self._cleanup_stored_futures()
            self._workers = dict()
            self._is_terminated = True
            self._stored_exception = None
            GlobalShutdown.deregister(self.shutdown)
            return True

    def _setup(self):
        GlobalShutdown.register(self.shutdown)
        GlobalShutdown.activate()

    def _spawn_filter_thread(self):
        thread_name = "asyncpal-{}-FilterThread".format(self._name)
        self._filter_thread = threading.Thread(name=thread_name,
                                               target=self._filter_tasks,
                                               daemon=False)
        self._filter_thread.start()

    def _spawn_workers(self, n=None):
        if n is None:
            n_workers = self._count_workers()
            n_free_workers = self._count_free_workers()
            n_tasks = self._count_pending_tasks()
            max_n = self._max_workers - n_workers
            n = max(0, min(max_n, n_tasks - n_free_workers))
        with self._workers_lock:
            for _ in range(n):
                self._monotonic_worker_count += 1
                worker = self._create_worker()
                self._workers[worker.worker_id] = worker
                worker.run()
        return n

    def _create_lock(self):
        return (threading.RLock()
                if self._worker_type == WorkerType.THREAD
                else self._mp_context.RLock())

    def _count_workers(self):
        with self._workers_lock:
            i = 0
            for worker in self._workers.values():
                if worker.is_alive():
                    i += 1
            return i

    def _count_busy_workers(self):
        with self._workers_lock:
            i = 0
            for worker in self._workers.values():
                if worker.is_busy():
                    i += 1
            return i

    def _count_free_workers(self):
        with self._workers_lock:
            i = 0
            for worker in self._workers.values():
                if not worker.is_busy():
                    i += 1
            return i

    def _count_pending_tasks(self):
        n = self._task_queue.qsize()
        if self._worker_type == WorkerType.PROCESS:
            n += self._mp_task_queue.qsize()
        return n

    def _ensure_pool_integrity(self):
        with self._vars_lock:
            if self._stored_exception is None:
                return
            if isinstance(self._stored_exception, errors.BrokenPoolError):
                raise self._stored_exception
            else:
                raise errors.BrokenPoolError from self._stored_exception

    @abstractmethod
    def _create_worker(self):
        pass

    def _submit_task(self, target, *args, **kwargs):
        self._join_inactive_workers()
        self._monotonic_task_count += 1
        task_id = self._monotonic_task_count
        future = Future(self, task_id)
        future.set_status(Status.PENDING)
        if self._worker_type == WorkerType.PROCESS:
            with self._futures_lock:
                self._stored_futures[task_id] = future
        task = (future, target, args, kwargs)
        self._task_queue.put(task)
        # ensure workers
        self._spawn_workers()
        return future

    def _map_lazy(self, target, iterable, buffer_size, timeout):
        countdown = misc.Countdown(timeout)
        buffer = deque()
        for i, args in enumerate(iterable):
            future = self._submit_task(target, *args)
            buffer.append(future)
            if i >= buffer_size - 1:
                future = buffer.popleft()
                new_timeout = countdown.check()
                yield future.collect(new_timeout)
        # consume buffer
        for future in buffer:
            new_timeout = countdown.check()
            yield future.collect(new_timeout)

    def _map_lazy_unordered(self, target, iterable, buffer_size, timeout):
        countdown = misc.Countdown(timeout)
        future_filter = FutureFilter()
        for i, args in enumerate(iterable):
            future = self._submit_task(target, *args)
            future_filter.put(future)
            if i >= buffer_size - 1:
                new_timeout = countdown.check()
                future = future_filter.get(timeout=new_timeout)
                yield future.collect()
        new_timeout = countdown.check()
        for future in future_filter.get_all(timeout=new_timeout):
            yield future.collect()

    def _map_lazy_chunked(self, target, iterable, chunk_size,
                          buffer_size, timeout):
        countdown = misc.Countdown(timeout)
        buffer = deque()
        for i, subtask in enumerate(misc.split_starmap_task(target,
                                                            iterable,
                                                            chunk_size=chunk_size)):
            future = self._submit_task(subtask)
            buffer.append(future)
            if i >= buffer_size - 1:
                future = buffer.popleft()
                new_timeout = countdown.check()
                for value in future.collect(new_timeout):
                    yield value
        # consume buffer
        for future in buffer:
            new_timeout = countdown.check()
            for value in future.collect(new_timeout):
                yield value

    def _map_lazy_chunked_unordered(self, target, iterable, chunk_size,
                                    buffer_size, timeout):
        countdown = misc.Countdown(timeout)
        future_filter = FutureFilter()
        for i, subtask in enumerate(misc.split_starmap_task(target,
                                                            iterable,
                                                            chunk_size=chunk_size)):
            future = self._submit_task(subtask)
            future_filter.put(future)
            if i >= buffer_size - 1:
                new_timeout = countdown.check()
                future = future_filter.get(timeout=new_timeout)
                for value in future.collect():
                    yield value
        new_timeout = countdown.check()
        for future in future_filter.get_all(timeout=new_timeout):
            for value in future.collect():
                yield value

    def _map_eager(self, target, iterable, keep_order, timeout):
        futures = [self._submit_task(target, *args) for args in iterable]
        return _collect_results(futures, keep_order=keep_order, timeout=timeout)

    def _map_eager_chunked(self, target, iterable, chunk_size,
                           keep_order, timeout):
        futures = list()
        for subtask in misc.split_starmap_task(target, iterable,
                                               chunk_size=chunk_size):
            future = self._submit_task(subtask)
            futures.append(future)
        return _collect_batched_results(futures, timeout=timeout,
                                        keep_order=keep_order)

    def _filter_tasks(self):
        while True:
            task = self._task_queue.get(block=True, timeout=None)
            if task is None:
                break
            future, target, args, kwargs = task
            if future.cancel_flag:
                future.set_status(Status.CANCELLED)
                with self._futures_lock:
                    del self._stored_futures[future.task_id]
                continue
            task = (future.task_id, target, args, kwargs)
            self._mp_task_queue.put(task, block=True, timeout=None)
            # delete references to objects as they might be holden
            # for too long because self._tasks_queue.get is a blocking call
            del task, future, target, args, kwargs

    def _on_worker_shutdown(self, worker_id):
        with self._workers_lock:
            try:
                worker = self._workers[worker_id]
            except KeyError as e:
                pass
            else:
                self._inactive_workers.append(worker)
                del self._workers[worker_id]
        if not self._is_closed.is_set():
            self._spawn_workers()

    def _on_worker_exception(self, worker_id, exc):
        with self._workers_lock:
            try:
                worker = self._workers[worker_id]
            except KeyError as e:
                pass
            else:
                self._inactive_workers.append(worker)
                del self._workers[worker_id]
        with self._vars_lock:
            if self._stored_exception is None:
                self._cancel_tasks()
                self._notify_workers_to_shutdown()
            self._stored_exception = exc
            self._is_broken = True

    def _notify_workers_to_shutdown(self):
        n = self._count_workers()
        task_queue = (self._task_queue
                      if self._worker_type == WorkerType.THREAD
                      else self._mp_task_queue)
        for _ in range(n):
            task_queue.put(None)

    def _cancel_tasks(self):  # called only once on shutdown !
        # drain the task queue and cancel futures
        tasks, futures = self._drain_task_queue()
        # drain the mp_task queue and cancel futures
        if self._worker_type == WorkerType.PROCESS:
            t, f = self._drain_mp_task_queue()
            tasks.extend(t)
            futures.extend(f)
        self._cancelled_tasks.extend(tasks)
        with self._futures_lock:
            for future in futures:
                future.set_status(Status.CANCELLED)
                try:
                    del self._stored_futures[future.task_id]
                except KeyError as e:
                    pass

    def _drain_task_queue(self):
        tasks = list()
        futures = list()
        # drain the task queue and cancel futures
        for task in misc.iterate_queue(self._task_queue):
            if task is None:
                continue
            future, target, args, kwargs = task
            futures.append(future)
            tasks.append((target, args, kwargs))
        return tasks, futures

    def _drain_mp_task_queue(self):
        tasks = list()
        task_ids = list()
        futures = list()
        for task in misc.iterate_queue(self._mp_task_queue):
            if task is None:
                continue
            task_id, target, args, kwargs = task
            task_ids.append(task_id)
            tasks.append((target, args, kwargs))
        with self._futures_lock:
            for task_id in task_ids:
                try:
                    future = self._stored_futures[task_id]
                except KeyError as e:
                    pass
                else:
                    futures.append(future)
        return tasks, futures

    def _join_inactive_workers(self):
        with self._workers_lock:
            for worker in self._inactive_workers:
                worker.join()
            self._inactive_workers = list()

    def _join_workers(self, timeout=None):
        with self._workers_lock:
            workers = tuple(self._workers.values())
        countdown = misc.Countdown(timeout)
        for worker in workers:
            if timeout is None:
                new_timeout = None
            else:
                new_timeout = countdown.check()
            if not worker.join(new_timeout):
                return False
        return True

    def _shutdown_message_thread(self):
        self._message_queue.put(None)
        self._message_thread.join()
        self._message_thread = None

    def _shutdown_filter_thread(self):
        self._task_queue.put(None)
        self._filter_thread.join()

    def _cleanup_task_queue(self):
        misc.drain_queue(self._task_queue)
        if self._worker_type == WorkerType.PROCESS:
            # spurious BrokenPipeError might happen while
            # trying to close the queue. A 1ms sleep fixes it !
            # See: https://github.com/python/cpython/issues/91185
            # See also: https://github.com/python/cpython/pull/31913
            time.sleep(0.001)
            # mp task queue is a multiprocessing Queue,
            # therefore to join it: .close() then .join_thread()
            self._mp_task_queue.close()
            self._mp_task_queue.join_thread()

    def _cleanup_stored_futures(self):
        with self._futures_lock:
            futures = self._stored_futures.values()
            self._stored_futures = dict()
        for future in futures:
            future.set_status(Status.CANCELLED)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    def __del__(self):
        self.shutdown()


def _collect_batched_results(futures, keep_order=True, timeout=None):
    for future in as_done(futures, keep_order=keep_order, timeout=timeout):
        for value in future.collect():
            yield value


def _collect_results(futures, keep_order=True, timeout=None):
    for future in as_done(futures, keep_order=keep_order, timeout=timeout):
        yield future.collect()


class GlobalShutdown:
    """Workaround to be able to use daemon threads and still gracefully shut down
    the pools at exit without having to use the atexit function"""
    _lock = threading.Lock()
    _handlers = list()
    _thread = None

    @classmethod
    def activate(cls):
        with cls._lock:
            if cls._thread is None:
                cls._thread = threading.Thread(target=global_shutdown_runner,
                                               daemon=False)
                cls._thread.start()

    @classmethod
    def is_active(cls):
        with cls._lock:
            return False if cls._thread is None else True

    @classmethod
    def join_thread(cls):
        """Join the globalshutdown thread"""
        with cls._lock:
            if cls._thread is not None:
                cls._thread.join()

    @classmethod
    def register(cls, handler):
        """Register a pool shutdown handler"""
        with cls._lock:
            cls._handlers.append(handler)

    @classmethod
    def deregister(cls, handler):
        """
        https://github.com/AxonFramework/AxonFramework/issues/2456
        https://english.stackexchange.com/questions/25931/unregister-vs-deregister
        """
        with cls._lock:
            cls._handlers = [h for h in cls._handlers if h != handler]

    @classmethod
    def count_handlers(cls):
        with cls._lock:
            return len(cls._handlers)

    @classmethod
    def get_handlers(cls):
        with cls._lock:
            return tuple(cls._handlers)


def global_shutdown_runner():
    """Join the main thread then run the shutdown handlers registered by pools.
    Note that the main thread can only be joined when it stops."""
    threading.main_thread().join(timeout=None)
    for handler in GlobalShutdown.get_handlers():
        handler()


atexit.register(GlobalShutdown.join_thread)
GlobalShutdown.activate()



import os
import sys
import unittest
import time
import itertools
from traceback import format_exception
from queue import Queue
from asyncpal import misc, errors
from asyncpal.pool import MP_CONTEXT
from tests import funcs


class TestGetChunksFunction(unittest.TestCase):

    def test_with_empty_iterable(self):
        r = tuple(misc.get_chunks(list(), chunk_size=1))
        self.assertEqual(tuple(), r)

    def test_with_chunk_size_equals_1(self):
        i = 0
        for chunk in misc.get_chunks([0, 1, 2, 3], chunk_size=1):
            with self.subTest(i):
                self.assertEqual(1, len(chunk))
                self.assertEqual(i, chunk[0])
            i += 1

    def test_with_chunk_size_equals_2(self):
        i = 0
        for chunk in misc.get_chunks([0, 1, 2, 3], chunk_size=2):
            with self.subTest(i):
                self.assertEqual(2, len(chunk))
                self.assertEqual(i, chunk[0])
                self.assertEqual(i+1, chunk[1])
            i += 2

    def test_with_chunk_size_equals_3(self):
        i = 0
        for chunk in misc.get_chunks([0, 1, 2, 3, 4, 5], chunk_size=3):
            with self.subTest(i):
                self.assertEqual(3, len(chunk))
                self.assertEqual(i, chunk[0])
                self.assertEqual(i+1, chunk[1])
                self.assertEqual(i+2, chunk[2])
            i += 3

    def test_with_too_big_chunk_size(self):
        i = 0
        for chunk in misc.get_chunks([0, 1, 2], chunk_size=5):
            with self.subTest(i):
                self.assertEqual(3, len(chunk))
                self.assertEqual((0, 1, 2), chunk)
                self.assertEqual(0, i)
            i += 1

    def test_with_odd_iterable_but_even_chunk_size(self):
        i = 0
        for chunk in misc.get_chunks([0, 1, 2], chunk_size=2):
            with self.subTest(i):
                if i == 0:
                    self.assertEqual((0, 1), chunk)
                elif i == 1:
                    self.assertEqual((2, ), chunk)
                else:
                    raise Exception
            i += 1


class TestEnsureArgsKwargs(unittest.TestCase):

    def test_without_arguments(self):
        r = misc.ensure_args_kwargs()
        expected = (tuple(), dict())
        self.assertEqual(expected, r)

    def test_with_args(self):
        r = misc.ensure_args_kwargs([0, 1], None)
        expected = ([0, 1], dict())
        self.assertEqual(expected, r)

    def test_with_kwargs(self):
        r = misc.ensure_args_kwargs(None, {"a": 0, "b": 1})
        expected = (tuple(), {"a": 0, "b": 1})
        self.assertEqual(expected, r)

    def test_with_args_and_kwargs(self):
        r = misc.ensure_args_kwargs([0, 1], {"a": 0, "b": 1})
        expected = ([0, 1], {"a": 0, "b": 1})
        self.assertEqual(expected, r)


class TestDrainQueue(unittest.TestCase):

    def test_drain_regular_queue(self):
        n = 10
        q = Queue()
        for x in range(n):
            q.put(x)
        self.assertEqual(n, q.qsize())
        misc.drain_queue(q)
        self.assertEqual(0, q.qsize())

    def test_drain_multiprocessing_queue(self):
        n = 10
        q = MP_CONTEXT.Queue()
        for x in range(n):
            q.put(x)
        time.sleep(0.01)  # give time to the background thread
        self.assertEqual(n, q.qsize())
        misc.drain_queue(q)
        time.sleep(0.01)
        self.assertEqual(0, q.qsize())


class TestIterateQueue(unittest.TestCase):

    def test_iterate_regular_queue(self):
        n = 10
        q = Queue()
        for x in range(n):
            q.put(x)
        r = tuple(misc.iterate_queue(q))
        self.assertEqual(tuple(range(n)), r)

    def test_iterate_multiprocessing_queue(self):
        n = 10
        q = MP_CONTEXT.Queue()
        for x in range(n):
            q.put(x)
        time.sleep(0.1)  # give time to the background thread
        r = tuple(misc.iterate_queue(q))
        self.assertEqual(tuple(range(n)), r)


class TestAddFunction(unittest.TestCase):

    def test(self):
        for x in range(10):
            with self.subTest(x):
                r = misc.add(x, x)
                self.assertEqual(x+x, r)


class TestCountdownClass(unittest.TestCase):

    def test_with_null_timeout(self):
        timeout = None
        timeout_updater = misc.Countdown(timeout)
        r = timeout_updater.check()
        self.assertIsNone(r)

    def test_with_zero_timeout(self):
        timeout = 0
        timeout_updater = misc.Countdown(timeout)
        r = timeout_updater.check()
        self.assertEqual(0, r)

    def test_with_expired_timeout(self):
        timeout = 0.01
        timeout_updater = misc.Countdown(timeout)
        time.sleep(0.01)
        r = timeout_updater.check()
        self.assertEqual(0, r)

    def test_with_unexpired_timeout(self):
        timeout = 0.01
        timeout_updater = misc.Countdown(timeout)
        r = timeout_updater.check()
        self.assertGreater(r, 0)


class TestSplitMapTask(unittest.TestCase):

    def test_without_chunk_size(self):
        result = list()
        subtasks = tuple(misc.split_map_task(funcs.square, range(10)))
        n = len(subtasks)
        for subtask in subtasks:
            result.extend(subtask())
        expected = tuple(map(funcs.square, range(10)))
        self.assertEqual(10, n)
        self.assertEqual(expected, tuple(result))

    def test_with_chunk_size(self):
        result = list()
        subtasks = tuple(misc.split_map_task(funcs.square, range(10),
                                             chunk_size=2))
        n = len(subtasks)
        for subtask in subtasks:
            result.extend(subtask())
        expected = tuple(map(funcs.square, range(10)))
        self.assertEqual(5, n)
        self.assertEqual(expected, tuple(result))


class TestSplitStarmapTask(unittest.TestCase):

    def test_without_chunk_size(self):
        result = list()
        it = zip(range(10), range(10))
        subtasks = tuple(misc.split_starmap_task(funcs.add, it))
        n = len(subtasks)
        for subtask in subtasks:
            result.extend(subtask())
        it = zip(range(10), range(10))
        expected = tuple(itertools.starmap(funcs.add, it))
        self.assertEqual(10, n)
        self.assertEqual(expected, tuple(result))

    def test_with_chunk_size(self):
        result = list()
        it = zip(range(10), range(10))
        subtasks = tuple(misc.split_starmap_task(funcs.add, it,
                                                 chunk_size=2))
        n = len(subtasks)
        for subtask in subtasks:
            result.extend(subtask())
        it = zip(range(10), range(10))
        expected = tuple(itertools.starmap(funcs.add, it))
        self.assertEqual(5, n)
        self.assertEqual(expected, tuple(result))


class TestCpuCount(unittest.TestCase):

    def test(self):
        if sys.version_info >= (3, 13):
            cpu_count = os.process_cpu_count()
        else:
            cpu_count = os.cpu_count()
        self.assertEqual(cpu_count, misc.get_cpu_count())


class TestFunctionGetRemoteTraceback(unittest.TestCase):

    def setUp(self):
        try:
            try:
                try:
                    try:
                        raise Exception("d")
                    except Exception as e:
                        raise Exception("c")
                except Exception as e:
                    raise Exception("b") from e
            except Exception as e:
                raise Exception("a") from e
        except Exception as e:
            self._exc = e
            self._exc_wrapper = misc.RemoteExceptionWrapper(e)

    def test(self):
        self.assertIsNone(misc.get_remote_traceback(self._exc))
        exc = self._exc_wrapper.unwrap()
        x = format_exception(type(self._exc), self._exc,
                             self._exc.__traceback__)
        x = "".join(x).strip()
        self.assertEqual(x, misc.get_remote_traceback(exc))


class TestRemoteExceptionWrapper(unittest.TestCase):

    def setUp(self):
        try:
            try:
                try:
                    try:
                        raise Exception("d")
                    except Exception as e:
                        raise Exception("c")
                except Exception as e:
                    raise Exception("b") from e
            except Exception as e:
                raise Exception("a") from e
        except Exception as e:
            self._exc = e
            self._exc_wrapper = misc.RemoteExceptionWrapper(e)

    def test_exc_chain_length(self):
        exc_chain = self._exc_wrapper.exc_chain
        self.assertEqual(4, len(exc_chain))

    def test_exc_chain(self):
        exc_chain = self._exc_wrapper.exc_chain
        exc_a, exc_a_has_cause = exc_chain[0]
        exc_b, exc_b_has_cause = exc_chain[1]
        exc_c, exc_c_has_cause = exc_chain[2]
        exc_d, exc_d_has_cause = exc_chain[3]
        self.assertEqual("a", str(exc_a))
        self.assertEqual("b", str(exc_b))
        self.assertEqual("c", str(exc_c))
        self.assertEqual("d", str(exc_d))
        self.assertTrue(exc_a_has_cause)
        self.assertTrue(exc_b_has_cause)
        self.assertFalse(exc_c_has_cause)
        self.assertFalse(exc_d_has_cause)

    def test_unwrap_method(self):
        exc = self._exc_wrapper.unwrap()
        self.assertEqual(str(self._exc), str(exc))
        self.assertEqual(str(self._exc.__cause__), str(exc.__cause__))
        self.assertEqual(str(self._exc.__cause__.__cause__), str(exc.__cause__.__cause__))
        self.assertEqual(str(self._exc.__cause__.__cause__.__context__),
                         str(exc.__cause__.__cause__.__context__))
        last_exc = exc.__cause__.__cause__.__context__
        self.assertIsInstance(last_exc.__context__, errors.RemoteTraceback)


if __name__ == "__main__":
    unittest.main()



import unittest


class TestImports(unittest.TestCase):

    def test_import_classes(self):
        try:
            # import classes
            from asyncpal import Pool
            from asyncpal import ThreadPool
            from asyncpal import ProcessPool
            from asyncpal import SingleThreadPool
            from asyncpal import SingleProcessPool
            from asyncpal import DualThreadPool
            from asyncpal import DualProcessPool
            from asyncpal import TripleThreadPool
            from asyncpal import TripleProcessPool
            from asyncpal import QuadThreadPool
            from asyncpal import QuadProcessPool
            from asyncpal import Future
            from asyncpal import FutureFilter
            from asyncpal import Countdown
            # import functions
            from asyncpal import as_done
            from asyncpal import wait
            from asyncpal import collect
            from asyncpal import split_map_task
            from asyncpal import split_starmap_task
            from asyncpal import get_chunks
            from asyncpal import get_remote_traceback
            # import constants
            from asyncpal import IDLE_TIMEOUT
            from asyncpal import MP_CONTEXT
            from asyncpal import LOGGER
            from asyncpal import WINDOWS_MAX_PROCESS_WORKERS
            # import errors
            from asyncpal import Error
            from asyncpal import BrokenPoolError
            from asyncpal import InitializerError
            from asyncpal import FinalizerError
            from asyncpal import InvalidStateError
            from asyncpal import CancelledError
        except ImportError:
            self.assertTrue(False)


if __name__ == "__main__":
    unittest.main()



import unittest
import time
from asyncpal import ThreadPool, errors, misc
from asyncpal.future import as_done, wait, collect, Future, FutureFilter
from tests import funcs


class TestResolveMethod(unittest.TestCase):

    def test_resolve_before_shutdown(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertEqual(3, future.collect())

    def test_resolve_after_shutdown(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2)
        self.assertEqual(3, future.collect())

    def test_resolve_with_pool_broken_by_initializer(self):
        with ThreadPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            future = pool.submit(funcs.add, 1, 2)
            with self.assertRaises(errors.CancelledError):
                future.collect()

    def test_resolve_with_pool_broken_by_finalizer(self):
        with ThreadPool(finalizer=funcs.divide, final_args=(1, 0)) as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertEqual(3, future.collect())
        self.assertTrue(True)

    def test_resolve_with_timeout(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2)
            self.assertEqual(3, future.collect(timeout=3))

    def test_resolve_with_too_short_timeout(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2, sleep=0.1)
            with self.assertRaises(TimeoutError):
                future.collect(timeout=0)  # 0 is The too short timeout ;)

    def test_resolve_with_task_exception(self):
        with ThreadPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            future = pool.submit(funcs.divide, 1, 0)
            with self.assertRaises(errors.CancelledError):
                future.collect()


class TestWaitMethod(unittest.TestCase):

    def test_wait_before_shutdown(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertTrue(future.wait())

    def test_wait_after_shutdown(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2)
            self.assertTrue(future.wait())

    def test_wait_with_pool_broken_by_initializer(self):
        with ThreadPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertTrue(future.wait())

    def test_wait_with_pool_broken_by_finalizer(self):
        with ThreadPool(finalizer=funcs.divide, final_args=(1, 0)) as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertTrue(future.wait())

    def test_wait_with_timeout(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2)
            self.assertTrue(future.wait(timeout=3))

    def test_wait_with_too_short_timeout(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, a=1, b=2, sleep=0.1)
            self.assertFalse(future.wait(timeout=0))  # 0 is The too short timeout ;)

    def test_wait_with_task_exception(self):
        with ThreadPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            future = pool.submit(funcs.divide, 1, 0)
            future.wait()
        self.assertTrue(True)


class TestCancelMethod(unittest.TestCase):

    def test(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 3, 4)
            time.sleep(0.001)
            future_1.cancel()
            future_2.cancel()
        self.assertFalse(future_1.is_cancelled)
        self.assertTrue(future_2.is_cancelled)


class TestCallbacks(unittest.TestCase):

    def test_callbacks_property(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            future.add_callbacks((dummy_callback_1, dummy_callback_2,
                                  dummy_callback_3))
            future.remove_callback(dummy_callback_3)
        expected = (dummy_callback_1, dummy_callback_2)
        self.assertEqual(expected, future.callbacks)

    def test_add_callback_early(self):
        cache = list()
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future.add_callback(cache.append)
            future.add_callback(cache.append)
            future.add_callbacks((cache.append, cache.append))
        expected = [future for _ in range(4)]
        self.assertEqual(expected, cache)

    def test_add_callback_late(self):
        cache = list()
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            future.collect()
        future.add_callback(cache.append)
        future.add_callback(cache.append)
        future.add_callbacks((cache.append, cache.append))
        expected = [future for _ in range(4)]
        self.assertEqual(expected, cache)

    def test_remove_callback(self):
        cache = list()
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2, sleep=0.1)
            # add callback
            future.add_callbacks((cache.append,
                                 cache.append,
                                 dummy_callback_1,
                                 dummy_callback_2,
                                 lambda f: cache.append(f)))
            # remove callbacks
            future.remove_callback(cache.append)
            future.remove_callbacks((dummy_callback_1,
                                    dummy_callback_2))
        expected = [future]
        self.assertEqual(expected, cache)


class TestProperties(unittest.TestCase):

    def test_pool_property(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertIs(pool, future.pool)

    def test_task_id_property(self):
        with ThreadPool() as pool:
            future_1 = pool.submit(funcs.add, 1, 2)
            future_2 = pool.submit(funcs.add, 1, 2)
            self.assertEqual(1, future_1.task_id)
            self.assertEqual(2, future_2.task_id)

    def test_pending_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertFalse(future_1.is_pending)
            self.assertTrue(future_2.is_pending)
            self.assertTrue(future_3.is_pending)
        self.assertFalse(future_1.is_pending)
        self.assertFalse(future_2.is_pending)
        self.assertFalse(future_3.is_pending)

    def test_running_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertTrue(future_1.is_running)
            self.assertFalse(future_2.is_running)
            self.assertFalse(future_3.is_running)
        self.assertFalse(future_1.is_running)
        self.assertFalse(future_2.is_running)
        self.assertFalse(future_3.is_running)

    def test_completed_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertFalse(future_1.is_completed)
            self.assertFalse(future_2.is_completed)
            self.assertFalse(future_3.is_completed)
            future_1.wait()
            future_2.wait()
            future_3.wait()
        self.assertTrue(future_1.is_completed)
        self.assertTrue(future_2.is_completed)
        self.assertTrue(future_3.is_completed)

    def test_failed_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.divide, 1, 0, sleep=0.1)
            self.assertFalse(future_1.is_failed)
            future_1.wait()
            self.assertTrue(future_1.is_failed)

    def test_cancelled_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            self.assertFalse(future_1.is_cancelled)
            self.assertFalse(future_2.is_cancelled)
        self.assertFalse(future_1.is_cancelled)
        self.assertTrue(future_2.is_cancelled)  # task_2 got cancelled as the pool shut down...

    def test_cancel_flag(self):
        with ThreadPool(max_workers=1) as pool:
            future = pool.submit(funcs.add, 1, 2)
            future.cancel()
            self.assertTrue(future.cancel_flag)

    def test_done_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.divide, 1, 0)  # ZeroDivisionError !
            future_2 = pool.submit(funcs.add, 1, 2)
            time.sleep(0.1)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertTrue(future_1.is_done)
            self.assertTrue(future_2.is_done)
            self.assertFalse(future_3.is_done)
        # failed (ZeroDivisionError), yet done
        self.assertTrue(future_1.is_failed)
        self.assertTrue(future_1.is_done)
        # completed, yet done
        self.assertTrue(future_2.is_completed)
        self.assertTrue(future_2.is_done)
        # cancelled (pool had to shut down quick), yet done
        self.assertTrue(future_3.is_cancelled)
        self.assertTrue(future_3.is_done)

    def test_result_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.divide, 1, 0)  # ZeroDivisionError !
            future_2 = pool.submit(funcs.add, 1, 2)
            time.sleep(0.1)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertIsNone(future_1.result)
            self.assertEqual(3, future_2.result)
            self.assertIsNone(future_3.result)
        self.assertIsNone(future_1.result)
        self.assertEqual(3, future_2.result)
        self.assertIsNone(future_3.result)

    def test_exception_property(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.divide, 1, 0)  # ZeroDivisionError !
            future_2 = pool.submit(funcs.add, 1, 2)
            time.sleep(0.1)
            future_3 = pool.submit(funcs.add, 1, 2)
            self.assertIsInstance(future_1.exception, ZeroDivisionError)
            self.assertIsNone(future_2.exception)
            self.assertIsNone(future_3.exception)
        self.assertIsNone(future_2.exception)
        self.assertIsNone(future_3.exception)


class TestTaskDuration(unittest.TestCase):

    def test_with_pending_task(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2, sleep=0.1)
            pending_duration, task_duration = future.duration
            self.assertEqual(0, pending_duration)
            self.assertEqual(0, task_duration)

    def test_with_completed_task(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            future.collect()
        pending_duration, task_duration = future.duration
        self.assertGreater(pending_duration, 0)
        self.assertGreater(task_duration, 0)

    def test_with_failed_task(self):
        with ThreadPool() as pool:
            future = pool.submit(funcs.divide, 1, 0)
            future.wait()
        pending_duration, task_duration = future.duration
        self.assertGreater(pending_duration, 0)
        self.assertGreater(task_duration, 0)

    def test_with_cancelled_task(self):
        with ThreadPool(max_workers=1) as pool:
            pool.submit(funcs.add, 1, 2, sleep=0.1)
            future = pool.submit(funcs.add, 1, 2)
        pending_duration, task_duration = future.duration
        self.assertGreater(pending_duration, 0)
        self.assertEqual(task_duration, 0)


class TestAsDoneFunction(unittest.TestCase):

    def test_with_order_maintained(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            it = as_done(futures, keep_order=True)  # keep_order defaults to False
            r = [future.collect() for future in it]
            expected = [1, 3, 5]
            self.assertEqual(expected, r)

    def test_without_order_maintained(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            it = as_done(futures, keep_order=False)  # keep_order defaults to False
            r = [future.collect() for future in it]
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_with_unexpired_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            it = as_done(futures, keep_order=True, timeout=1)
            r = [future.collect() for future in it]
            expected = [1, 3, 5]
            self.assertEqual(expected, r)

    def test_with_expired_timeout(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.1)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3, sleep=0.1)
            futures = (future_1, future_2, future_3)
            with self.assertRaises(TimeoutError):
                tuple(as_done(futures, keep_order=True, timeout=0.05))


class TestWaitFunction(unittest.TestCase):

    def test_wait_without_timeout(self):
        with ThreadPool() as pool:
            future_1 = pool.submit(funcs.add, 1, 2)
            future_2 = pool.submit(funcs.add, 3, 4)
            futures = (future_1, future_2)
            r = wait(futures)  # timeout defaults to None
            self.assertTrue(r)

    def test_wait_with_expired_timeout(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.1)
            futures = (future_1, future_2)
            r = wait(futures, timeout=0.1)
            self.assertFalse(r)

    def test_wait_with_unexpired_timeout(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.1)
            futures = (future_1, future_2)
            r = wait(futures, timeout=5)
            self.assertTrue(r)


class TestResolveFunction(unittest.TestCase):

    def test_resolve_without_exception(self):
        with ThreadPool() as pool:
            future_1 = pool.submit(funcs.add, 1, 2)
            future_2 = pool.submit(funcs.add, 3, 4)
            futures = (future_1, future_2)
            r = collect(futures)
            expected = (3, 7)
            self.assertEqual(expected, r)

    def test_resolve_with_exception(self):
        with ThreadPool() as pool:
            future_1 = pool.submit(funcs.divide, 2, 1)
            future_2 = pool.submit(funcs.divide, 1, 0)
            futures = (future_1, future_2)
            with self.assertRaises(ZeroDivisionError):
                collect(futures)

    def test_resolve_with_expired_timeout(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.1)
            futures = (future_1, future_2)
            with self.assertRaises(TimeoutError):
                collect(futures, timeout=0.1)

    def test_resolve_with_unexpired_timeout(self):
        with ThreadPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.1)
            futures = (future_1, future_2)
            r = collect(futures, timeout=5)
            expected = (3, 7)
            self.assertEqual(expected, r)


class TestFuture(unittest.TestCase):
    # Note that the Future class is already indirectly
    # tested throughout other tests

    def test(self):
        with ThreadPool() as pool:
            future = Future(pool, 1)
            future.set_result(42)
            self.assertIs(pool, future.pool)
            self.assertEqual(1, future.task_id)
            self.assertEqual(42, future.collect())


class TestFutureFilter(unittest.TestCase):

    def test_put_method(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            future_filter = FutureFilter()
            future_filter.put(future_1)
            future_filter.put(future_2)
            future_filter.put(future_3)
            r = drain_future_filter(future_filter)
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_populate_method(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            future_filter = FutureFilter()
            future_filter.populate((future_1, future_2, future_3))
            r = drain_future_filter(future_filter)
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_get_method_with_blocking(self):
        with ThreadPool(max_workers=4) as pool:
            future_filter = FutureFilter()
            with self.subTest():
                self.assertIsNone(future_filter.get(block=True))  # blocks by default
            with self.subTest():
                future = pool.submit(funcs.add, 0, 1)
                future_filter.put(future)
                f = future_filter.get(block=True)
                self.assertEqual(1, f.collect())
            with self.subTest():
                future = pool.submit(funcs.add, 1, 2, sleep=0.1)
                future_filter.put(future)
                # note that FutureFilter.get blocks by default
                # (block=True, timeout=None)
                f = future_filter.get(block=True)
                expected = 3
                self.assertEqual(expected, f.collect())

    def test_get_method_without_blocking(self):
        with ThreadPool(max_workers=4) as pool:
            future_filter = FutureFilter()
            with self.subTest():
                self.assertIsNone(future_filter.get(block=False))
            with self.subTest():
                future = pool.submit(funcs.add, 0, 1)
                future_filter.put(future)
                f = future_filter.get(block=False)
                self.assertEqual(1, f.collect())
            with self.subTest():
                future = pool.submit(funcs.add, 1, 2, sleep=0.1)
                future_filter.put(future)
                f = future_filter.get(block=False)
                self.assertIsNone(f)

    def test_get_method_without_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            r = drain_future_filter(future_filter)
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_get_method_with_expired_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            with self.assertRaises(TimeoutError):
                drain_future_filter(future_filter, timeout=0.02)

    def test_get_method_with_unexpired_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            r = drain_future_filter(future_filter, timeout=1)
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_get_all_method_without_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            r = [future.collect() for future in future_filter.get_all()]
            expected = [5, 3, 1]
            self.assertEqual(expected, r)

    def test_get_all_method_with_expired_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            with self.assertRaises(TimeoutError):
                [future.collect() for future
                 in future_filter.get_all(timeout=0.02)]

    def test_get_all_method_with_unexpired_timeout(self):
        with ThreadPool(max_workers=4) as pool:
            future_1 = pool.submit(funcs.add, 0, 1, sleep=0.5)
            future_2 = pool.submit(funcs.add, 1, 2, sleep=0.1)
            future_3 = pool.submit(funcs.add, 2, 3)
            futures = (future_1, future_2, future_3)
            future_filter = FutureFilter(futures)
            r = [future.collect() for future
                 in future_filter.get_all(timeout=1)]
            expected = [5, 3, 1]
            self.assertEqual(expected, r)


def dummy_callback_1(future):
    pass


def dummy_callback_2(future):
    pass


def dummy_callback_3(future):
    pass


def drain_future_filter(future_filter, block=True, timeout=None):
    countdown = misc.Countdown(timeout)
    n = len(future_filter.futures)
    result = list()
    for _ in range(n):
        new_timeout = countdown.check()
        future = future_filter.get(block=block, timeout=new_timeout)
        result.append(future.collect())
    return result


if __name__ == "__main__":
    unittest.main()



import unittest
import time
import itertools
from asyncpal import errors
from tests import funcs
from asyncpal import (ProcessPool, SingleProcessPool,
                      DualProcessPool, TripleProcessPool,
                      QuadProcessPool)


class TestWorkers(unittest.TestCase):

    def test_worker(self):
        with ProcessPool() as pool:
            pool.test()
        self.assertTrue(True)

    def test_worker_after_shutdown(self):
        with ProcessPool() as pool:
            pass
        with self.assertRaises(RuntimeError):
            pool.test()

    def test_worker_with_broken_pool(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.BrokenPoolError):
                pool.test()
            worker_exception = funcs.get_worker_exception(pool)
            first_cause = worker_exception.__cause__
            self.assertIsInstance(first_cause, ZeroDivisionError)


class TestBrokenPool(unittest.TestCase):

    def test_before_shutdown_with_pool_broken_by_initializer(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.InitializerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()
            worker_exception = funcs.get_worker_exception(pool)
            self.assertIsInstance(worker_exception, errors.InitializerError)
            self.assertIsInstance(worker_exception, errors.BrokenPoolError)
            first_cause = worker_exception.__cause__
            self.assertIsInstance(first_cause, ZeroDivisionError)
            self.assertTrue(pool.is_broken)

    def test_after_shutdown_with_pool_broken_by_initializer(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.InitializerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()
        with self.assertRaises(RuntimeError):
            pool.check()

    def test_before_shutdown_with_pool_broken_by_finalizer(self):
        # idle_timeout is set to 0 so the worker will quickly
        # return, giving chance to finalizer to run before 'check()'
        with ProcessPool(finalizer=funcs.divide, final_args=(1, 0),
                         idle_timeout=0) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.FinalizerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()
            worker_exception = funcs.get_worker_exception(pool)
            self.assertIsInstance(worker_exception, errors.FinalizerError)
            self.assertIsInstance(worker_exception, errors.BrokenPoolError)
            first_cause = worker_exception.__cause__
            self.assertIsInstance(first_cause, ZeroDivisionError)
            self.assertTrue(pool.is_broken)

    def test_after_shutdown_with_pool_broken_by_finalizer(self):
        with ProcessPool(finalizer=funcs.divide, final_args=(1, 0)) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.FinalizerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()
        with self.assertRaises(RuntimeError):
            pool.check()

    def test_with_pool_broken_by_initializer_missing_args_error(self):
        with ProcessPool(initializer=funcs.divide) as pool:
            pool.spawn_workers(1)
            pool.join()
            worker_exception = funcs.get_worker_exception(pool)
            first_cause = worker_exception.__cause__
            self.assertIsInstance(first_cause, TypeError)
            self.assertTrue(pool.is_broken)


class TestCheckMethod(unittest.TestCase):

    def test_check_before_shutdown(self):
        with ProcessPool() as pool:
            pool.check()
        self.assertTrue(True)

    def test_check_after_shutdown(self):
        with ProcessPool() as pool:
            pass
        with self.assertRaises(RuntimeError):
            pool.check()

    def test_check_before_shutdown_with_pool_broken_by_initializer(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0), idle_timeout=1) as pool:
            with self.assertRaises(errors.CancelledError):
                # I could've put 'pool.spawn_workers(1)' here
                # but 'pool.check()' risks to complete before
                # the worker returns
                pool.submit(funcs.add, 1, 2).collect()
            with self.assertRaises(errors.InitializerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()

    def test_check_after_shutdown_with_pool_broken_by_initializer(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            # I could've put 'pool.submit(funcs.add, 1, 2).resolve()' here
            # instead of 'pool.spawn_workers(1)', but it is not needed
            # as 'pool.check()' will be run after the shutdown of the pool,
            # which guarantees that the worker would've already returned.
            pool.spawn_workers(1)
        with self.assertRaises(RuntimeError):
            pool.check()

    def test_check_before_shutdown_with_pool_broken_by_finalizer(self):
        with ProcessPool(finalizer=funcs.divide, final_args=(1, 0)) as pool:
            pool.spawn_workers(1)
            pool.join()
            with self.assertRaises(errors.FinalizerError):
                pool.check()
            with self.assertRaises(errors.BrokenPoolError):
                pool.check()

    def test_check_after_shutdown_with_pool_broken_by_finalizer(self):
        with ProcessPool(finalizer=funcs.divide, final_args=(1, 0)) as pool:
            pool.spawn_workers(1)
        with self.assertRaises(RuntimeError):
            pool.check()


class TestRunMethod(unittest.TestCase):

    def test_submit_to_pool(self):
        with ProcessPool() as pool:
            r = pool.run(funcs.add, 1, 2)
            self.assertEqual(3, r)

    def test_submit_to_broken_pool(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            with self.assertRaises(errors.CancelledError):
                pool.run(funcs.add, 1, 2)

    def test_submit_to_closed_pool(self):
        with ProcessPool() as pool:
            pass
        with self.assertRaises(RuntimeError):
            pool.run(funcs.add, 1, 2)

    def test_submit_to_broken_and_closed_pool(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            pool.spawn_workers(1)
        with self.assertRaises(RuntimeError):
            pool.run(funcs.add, 1, 2)


class TestSubmitMethod(unittest.TestCase):

    def test_submit_to_pool(self):
        with ProcessPool() as pool:
            future = pool.submit(funcs.add, 1, 2)
            self.assertEqual(3, future.collect())

    def test_submit_to_broken_pool(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            future = pool.submit(funcs.add, 1, 2)
            with self.assertRaises(errors.CancelledError):
                future.collect()

    def test_submit_to_closed_pool(self):
        with ProcessPool() as pool:
            pass
        with self.assertRaises(RuntimeError):
            pool.submit(funcs.add, 1, 2)

    def test_submit_to_broken_and_closed_pool(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            pool.spawn_workers(1)
        with self.assertRaises(RuntimeError):
            pool.submit(funcs.add, 1, 2)


class TestProperties(unittest.TestCase):

    def test_closed_property(self):
        with ProcessPool() as pool:
            self.assertFalse(pool.is_closed)
        self.assertTrue(pool.is_closed)

    def test_terminated_property(self):
        with ProcessPool() as pool:
            self.assertFalse(pool.is_terminated)
        self.assertTrue(pool.is_terminated)

    def test_broken_property(self):
        with ProcessPool(initializer=funcs.divide, init_args=(1, 0)) as pool:
            self.assertFalse(pool.is_broken)
            pool.spawn_workers(1)
        self.assertTrue(pool.is_broken)


class TestLazyMapMethod(unittest.TestCase):

    def test_with_defaults(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map(funcs.square, range(10)))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 2)

    def test_with_custom_chunk_size(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map(funcs.square, range(10), chunk_size=5))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 2)

    def test_with_buffer_size(self):
        # as processes aren't as reactive as threads, one extra process
        # might be spawn, that's why instead of 3 (the buffer size),
        # I've put 4 in self.assertLessEqual(pool.count_workers(), 4)
        # and increased the max_workers from 4 to 7
        with ProcessPool(max_workers=7) as pool:
            r = tuple(pool.map(funcs.square, range(10), chunk_size=1,
                               buffer_size=3))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 4)

    def test_with_unordered_result(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map(funcs.square, range(3),
                               (0.5, 0.2, 0),
                               keep_order=False, buffer_size=3))
            expected = tuple(map(funcs.square, (2, 1, 0)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 3)

    def test_with_unexpired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map(funcs.square, range(10), timeout=1))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)

    def test_with_expired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            with self.assertRaises(TimeoutError):
                tuple(pool.map(funcs.square,
                               range(5),  # square numbers from 0 to 4
                               [0.02 for _ in range(5)],  # sleep for 0.01 s
                               timeout=0.01))


class TestEagerMapMethod(unittest.TestCase):

    def test_with_defaults(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map_all(funcs.square, range(10)))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)
            self.assertEqual(4, pool.count_workers())

    def test_with_custom_chunk_size(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map_all(funcs.square, range(10), chunk_size=5))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 2)

    def test_with_unordered_result(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map_all(funcs.square, range(3),
                                   (0.6, 0.3, 0),
                                   keep_order=False))
            expected = tuple(map(funcs.square, (2, 1, 0)))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 3)

    def test_with_unexpired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.map_all(funcs.square, range(10), timeout=5))
            expected = tuple(map(funcs.square, range(10)))
            self.assertEqual(expected, r)

    def test_with_expired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            with self.assertRaises(TimeoutError):
                tuple(pool.map_all(funcs.square,
                                   range(5),  # square numbers from 0 to 4
                                   [0.02 for _ in range(5)],  # sleep for 0.01 s
                                   timeout=0.01))


class TestLazyStarmapMethod(unittest.TestCase):

    def test_with_defaults(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap(funcs.add, zip(range(10), range(10))))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)
            self.assertEqual(1, pool.count_workers())

    def test_with_custom_chunk_size(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap(funcs.add, zip(range(10), range(10)), chunk_size=5))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 2)

    def test_with_buffer_size(self):
        # as processes aren't as reactive as threads, one extra process
        # might be spawn, that's why instead of 3 (the buffer size),
        # I've put 4 in self.assertLessEqual(pool.count_workers(), 4)
        # and increased the max_workers from 4 to 7
        with ProcessPool(max_workers=7) as pool:
            r = tuple(pool.starmap(funcs.add, zip(range(10), range(10)),
                                   chunk_size=1, buffer_size=3))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 4)

    def test_with_unordered_result(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap(funcs.add, zip(range(3), range(3),
                                                  (0.5, 0.2, 0)),
                                   keep_order=False, buffer_size=3))
            expected = tuple(itertools.starmap(funcs.add,
                                               [(2, 2), (1, 1), (0, 0)]))
            self.assertEqual(expected, r)
            self.assertEqual(3, pool.count_workers())

    def test_with_unexpired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap(funcs.add, zip(range(10), range(10)), timeout=1))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)

    def test_with_expired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            with self.assertRaises(TimeoutError):
                tuple(pool.starmap(funcs.add,
                                   zip(range(5), range(5),
                                       [0.02 for _ in range(5)]),
                                   timeout=0.01))


class TestEagerStarmapMethod(unittest.TestCase):

    def test_with_defaults(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap_all(funcs.add, zip(range(10), range(10))))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)
            self.assertEqual(4, pool.count_workers())

    def test_with_custom_chunk_size(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap_all(funcs.add, zip(range(10), range(10)), chunk_size=5))
            expected = tuple(itertools.starmap(funcs.add, zip(range(10), range(10))))
            self.assertEqual(expected, r)
            self.assertLessEqual(pool.count_workers(), 2)

    def test_with_unordered_result(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap_all(funcs.add, zip(range(3), range(3),
                                                      (0.5, 0.2, 0)),
                                       keep_order=False))
            expected = tuple(itertools.starmap(funcs.add,
                                               [(2, 2), (1, 1), (0, 0)]))
            self.assertEqual(expected, r)
            self.assertEqual(3, pool.count_workers())

    def test_with_unexpired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            r = tuple(pool.starmap_all(funcs.add, zip(range(10),
                                                      range(10)), timeout=1))
            expected = tuple(itertools.starmap(funcs.add,
                                               zip(range(10), range(10))))
            self.assertEqual(expected, r)

    def test_with_expired_timeout(self):
        with ProcessPool(max_workers=4) as pool:
            with self.assertRaises(TimeoutError):
                tuple(pool.starmap_all(funcs.add,
                                       zip(range(5), range(5),
                                           [0.02 for _ in range(5)]),
                                       timeout=0.01))


class TestSpawnWorkersMethod(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=4) as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_workers(3)
            self.assertEqual(3, pool.count_workers())


class TestCountWorkersMethod(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=4) as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_workers(1)
            self.assertEqual(1, pool.count_workers())
            pool.join()
            self.assertEqual(0, pool.count_workers())


class TestCountBusyWorkersMethod(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=1) as pool:
            self.assertEqual(0, pool.count_busy_workers())
            pool.submit(funcs.add, 1, 2, sleep=1)
            time.sleep(0.5)
            self.assertEqual(1, pool.count_busy_workers())
            pool.join()
            self.assertEqual(0, pool.count_busy_workers())


class TestCountPendingTasksMethod(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=1) as pool:
            self.assertEqual(0, pool.count_pending_tasks())
            pool.submit(funcs.add, 1, 2, sleep=0.01)
            pool.submit(funcs.add, 1, 2)
            pool.submit(funcs.add, 1, 2)
            self.assertLessEqual(pool.count_pending_tasks(), 3)


class TestCancelledTasksProperty(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=1) as pool:
            self.assertEqual(0, len(pool.cancelled_tasks))
            pool.submit(funcs.add, 1, 2, sleep=0.01)
            pool.submit(funcs.add, 1, 2)  # will be cancelled soon
            pool.submit(funcs.add, 1, 2)  # will be cancelled soon
        self.assertLessEqual(len(pool.cancelled_tasks), 3)


class TestSpawnMaxWorkers(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=4) as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_max_workers()
            self.assertEqual(4, pool.count_workers())
        self.assertEqual(0, pool.count_workers())


class TestJoinPool(unittest.TestCase):

    def test_join(self):
        with ProcessPool(max_workers=1, idle_timeout=60) as pool:
            instant_a = time.monotonic()
            pool.spawn_workers(1)
            self.assertTrue(pool.join())
            # t should be <= 1 second
            # despite the 60 seconds idle_timeout
            t = time.monotonic() - instant_a
            self.assertLessEqual(t, 1)

    def test_join_with_unexpired_timeout(self):
        with ProcessPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.01)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.01)
            self.assertTrue(pool.join(5))
            self.assertTrue(3, future_1.result)
            self.assertTrue(7, future_2.result)

    def test_join_with_expired_timeout(self):
        with ProcessPool(max_workers=1) as pool:
            future_1 = pool.submit(funcs.add, 1, 2, sleep=0.01)
            future_2 = pool.submit(funcs.add, 3, 4, sleep=0.01)
            self.assertFalse(pool.join(0.01))
            self.assertTrue(3, future_1.result)
            self.assertIsNone(future_2.result)


class TestMaxTasksByWorker(unittest.TestCase):

    def test(self):
        with ProcessPool(max_workers=1, max_tasks_per_worker=3, idle_timeout=60) as pool:
            pool.run(funcs.add, 1, 2)
            pool.run(funcs.add, 3, 4)
            pool.run(funcs.add, 5, 6)
            time.sleep(0.01)
            self.assertEqual(0, pool.count_workers())


class TestSizedPools(unittest.TestCase):

    def test_single_process_pool(self):
        with SingleProcessPool() as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_max_workers()
            self.assertEqual(1, pool.count_workers())
        self.assertEqual(0, pool.count_workers())

    def test_dual_process_pool(self):
        with DualProcessPool() as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_max_workers()
            self.assertEqual(2, pool.count_workers())
        self.assertEqual(0, pool.count_workers())

    def test_triple_process_pool(self):
        with TripleProcessPool() as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_max_workers()
            self.assertEqual(3, pool.count_workers())
        self.assertEqual(0, pool.count_workers())

    def test_quad_process_pool(self):
        with QuadProcessPool() as pool:
            self.assertEqual(0, pool.count_workers())
            pool.spawn_max_workers()
            self.assertEqual(4, pool.count_workers())
        self.assertEqual(0, pool.count_workers())
        
        
if __name__ == "__main__":
    unittest.main()



"""ThreadPool, SingleThreadPool, DualThreadPool,
TripleThreadPool, and QuadThreadPool are defined here"""
import queue
from asyncpal import misc
from asyncpal.worker.threadworker import ThreadWorker
from asyncpal.pool import Pool, WorkerType, IDLE_TIMEOUT


__all__ = ["ThreadPool", "SingleThreadPool", "DualThreadPool",
           "TripleThreadPool", "QuadThreadPool"]


class ThreadPool(Pool):
    """The ThreadPool class for  preemptive concurrency."""
    def __init__(self, max_workers=None, *, name="ThreadPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None):
        """
        Initialization.

        [param]
        - max_workers: the maximum number of workers. Defaults to CPU count + 5
        - name: the name of the pool. Defaults to the class name
        - idle_timeout: None or a timeout value in seconds.
            The idle timeout tells how much time an inactive worker
            can sleep before it closes. This helps the pool to shrink
            when there isn't much of tasks.
            If you set None, the pool will never shrink. each worker before it closes
        - initializer: a function that will get called at the start of each worker
        - init_args: arguments (list) to pass to the initializer
        - init_kwargs: keyword arguments (dict) to pass to the initializer
        - finalizer: a function that will get called when the worker is going to close
        - final_args: arguments (list) to pass to the finalizer
        - final_kwargs: keyword arguments (dict) to pass to the finalizer
        - max_tasks_per_worker: Maximum number of tasks a worker is allowed to do
            before it closes.
        """
        if max_workers is None or max_workers <= 0:
            x = misc.get_cpu_count() + 5
            max_workers = x if x else 1
        worker_type = WorkerType.THREAD
        task_queue = queue.SimpleQueue()
        super().__init__(worker_type, max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer,
                         init_args=init_args, init_kwargs=init_kwargs,
                         finalizer=finalizer, final_args=final_args,
                         final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         task_queue=task_queue)

    def _create_worker(self):
        worker_id = self._monotonic_worker_count
        worker_name = "asyncpal-{}-thread-worker-{}".format(self._name, worker_id)
        with self._vars_lock:
            worker = ThreadWorker(worker_id, worker_name, self._task_queue,
                                  self._idle_timeout, self._initializer,
                                  self._init_args, self._init_kwargs,
                                  self._finalizer, self._final_args,
                                  self._final_kwargs, self._max_tasks_per_worker,
                                  self._on_worker_shutdown, self._on_worker_exception)
            return worker


class SingleThreadPool(ThreadPool):
    """Fixed-size thread pool. This pool can spawn only 1 worker"""
    def __init__(self, *, name="SingleThreadPool", idle_timeout=IDLE_TIMEOUT,
                 initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None):
        max_workers = 1
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer,
                         init_args=init_args, init_kwargs=init_kwargs,
                         finalizer=finalizer, final_args=final_args,
                         final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker)


class DualThreadPool(ThreadPool):
    """Fixed-size thread pool. This pool can spawn up to 2 workers"""
    def __init__(self, *, name="DualThreadPool", idle_timeout=IDLE_TIMEOUT,
                 initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None):
        max_workers = 2
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer,
                         init_args=init_args, init_kwargs=init_kwargs,
                         finalizer=finalizer, final_args=final_args,
                         final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker)


class TripleThreadPool(ThreadPool):
    """Fixed-size thread pool. This pool can spawn up to 3 workers"""
    def __init__(self, *, name="TripleThreadPool", idle_timeout=IDLE_TIMEOUT,
                 initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None):
        max_workers = 3
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer,
                         init_args=init_args, init_kwargs=init_kwargs,
                         finalizer=finalizer, final_args=final_args,
                         final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker)


class QuadThreadPool(ThreadPool):
    """Fixed-size thread pool. This pool can spawn up to 4 workers"""
    def __init__(self, *, name="QuadThreadPool", idle_timeout=IDLE_TIMEOUT,
                 initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None):
        max_workers = 4
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer,
                         init_args=init_args, init_kwargs=init_kwargs,
                         finalizer=finalizer, final_args=final_args,
                         final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker)



from abc import ABC, abstractmethod
from enum import Enum, unique

__all__ = []

@unique
class WorkerType(Enum):
    THREAD = 1
    PROCESS = 2


class Worker(ABC):
    def __init__(self, worker_type, worker_id, worker_name,
                 task_queue, idle_timeout,
                 initializer, init_args, init_kwargs,
                 finalizer, final_args, final_kwargs,
                 max_tasks_per_worker):
        self._worker_type = worker_type
        self._worker_id = worker_id
        self._worker_name = worker_name
        self._task_queue = task_queue
        self._idle_timeout = idle_timeout
        self._initializer = initializer
        self._init_args = init_args
        self._init_kwargs = init_kwargs
        self._finalizer = finalizer
        self._final_args = final_args
        self._final_kwargs = final_kwargs
        self._max_tasks_per_worker = max_tasks_per_worker

    @property
    def worker_type(self):
        return self._worker_type

    @property
    def worker_id(self):
        return self._worker_id

    @property
    def worker_name(self):
        return self._worker_name

    @property
    def task_queue(self):
        return self._task_queue

    @property
    def idle_timeout(self):
        return self._idle_timeout

    @property
    def initializer(self):
        return self._initializer

    @property
    def init_args(self):
        return self._init_args

    @property
    def init_kwargs(self):
        return self._init_kwargs

    @property
    def finalizer(self):
        return self._finalizer

    @property
    def final_args(self):
        return self._final_args

    @property
    def final_kwargs(self):
        return self._final_kwargs

    @property
    def max_tasks_per_worker(self):
        return self._max_tasks_per_worker

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def is_alive(self):
        pass

    @abstractmethod
    def is_busy(self):
        pass

    @abstractmethod
    def join(self, timeout=None):
        pass



"""In this module are defined the `Future` and the `FutureFilter`
classes as well as the `wait`, `collect`, and `as_done` functions"""
import functools
import time
import threading
from enum import Enum, unique
from queue import SimpleQueue, Empty as EmptyQueue
from asyncpal import errors, misc


__all__ = ["Future", "FutureFilter", "Status", "wait", "collect", "as_done"]


def wait(futures, timeout=None):
    """
    Wait (blocking) for futures to get done (completed, failed, or cancelled).

    [param]
    - futures: sequence of Future objects
    - timeout: None or a timeout value (int or float) in seconds.

    [return]
    Returns True if all Futures are done in the provided timeout range.
    """
    countdown = misc.Countdown(timeout)
    for future in futures:
        if timeout is None:
            new_timeout = None
        else:
            new_timeout = countdown.check()
        if not future.wait(new_timeout):
            return False
    return True


def collect(futures, timeout=None):
    """
    Collect the results of a sequence of futures.
    Beware, remote exceptions as well as CancelledError,
    or TimeoutError exceptions might be raised.

    [param]
    - futures: sequence of Future objects
    - timeout: None or a timeout value (int or float) in seconds.

    [return]
    Returns a tuple containing the ordered results
    collected from all the futures

    [except]
    - TimeoutError: raised when timeout expires
    - CancelledError: raised when a task got cancelled
    """
    result = list()
    countdown = misc.Countdown(timeout)
    for future in futures:
        if timeout is None:
            new_timeout = None
        else:
            new_timeout = countdown.check()
        result.append(future.collect(new_timeout))
    return tuple(result)


def as_done(futures, keep_order=False, timeout=None):
    """
    Yields futures iteratively as they are done

    [param]
    - futures: sequence of futures
    - keep_order: boolean to tell whether the results should be ordered or not
    - timeout: None or a timeout value (int or float) in seconds.

    [except]
    - TimeoutError: raised when timeout expires
    """
    if keep_order:
        countdown = misc.Countdown(timeout)
        for future in futures:
            if timeout is None:
                new_timeout = None
            else:
                new_timeout = countdown.check()
            if not future.wait(new_timeout):
                raise TimeoutError
            yield future
    else:
        future_filter = FutureFilter(futures)
        yield from future_filter.get_all(block=True, timeout=timeout)


class FutureFilter:
    """
    This is a filter. First, you populate it with Future
    objects, then you retrieve futures as they are done.
    You can also retrieve futures while feeding the filter.
    The original order of the futures doesn't matter here.

    This class is used by the `as_done` function when its
    `keep_order` option is set to False.
    """

    def __init__(self, futures=None):
        self._futures = list()
        self._lock = threading.RLock()
        self._queue = SimpleQueue()
        self._n = len(self._futures)
        if futures:
            self.populate(futures)

    @property
    def futures(self):
        with self._lock:
            return tuple(self._futures)

    def put(self, future):
        """
        Add a future to the filter

        [param]
        - future: Future object
        """
        with self._lock:
            self._futures.append(future)
            future.add_callback(functools.partial(self._queue.put))
            self._n += 1

    def populate(self, futures):
        """
        Populate the filter

        [param]
        - futures: sequence of Future object
        """
        with self._lock:
            self._futures.extend(futures)
            self._n += len(futures)
        for future in futures:
            future.add_callback(functools.partial(self._queue.put))

    def get(self, block=True, timeout=None):
        """
        Retrieve just one future that is done.

        [param]
        - block: boolean to block until one future is done
        - timeout: None or a timeout value (int or float) in seconds.

        [return]
        Returns a future object that is done
        """
        with self._lock:
            if self._n == 0:
                return
        future = self._get_from_queue(block, timeout)
        return future

    def get_all(self, block=True, timeout=None):
        """
        Yield futures as they are done until the filter is empty

        [param]
        - block: boolean to block until one future is done
        - timeout: None or a timeout value (int or float) in seconds.

        [yield]
        Yield a future object that is done until the filter is empty
        """
        countdown = misc.Countdown(timeout)
        while True:
            with self._lock:
                if self._n == 0:
                    break
            new_timeout = countdown.check()
            future = self._get_from_queue(block, new_timeout)
            yield future

    def _get_from_queue(self, block, timeout):
        try:
            future = self._queue.get(block=block, timeout=timeout)
        except EmptyQueue as e:
            if block:
                raise TimeoutError
            else:
                return
        with self._lock:
            self._n -= 1
        return future


class Future:
    """Future class"""

    def __init__(self, pool, task_id):
        """This class shouldn't be instantiated by the programmer"""
        self._pool = pool
        self._task_id = task_id
        self._on_done_event = threading.Event()
        self._status = None
        self._is_done = False
        self._result = self._duration = self._exception = None
        self._callbacks = list()
        self._mutex = threading.RLock()
        self._pending_duration = self._task_duration = 0
        self._status_to_time = dict()
        self._cancel_flag = False

    @property
    def pool(self):
        return self._pool

    @property
    def task_id(self):
        return self._task_id

    @property
    def is_pending(self):
        with self._mutex:
            return True if self._status == Status.PENDING else False

    @property
    def is_running(self):
        with self._mutex:
            return True if self._status == Status.RUNNING else False

    @property
    def is_completed(self):
        with self._mutex:
            return True if self._status == Status.COMPLETED else False

    @property
    def is_failed(self):
        with self._mutex:
            return True if self._status == Status.FAILED else False

    @property
    def is_cancelled(self):
        with self._mutex:
            return True if self._status == Status.CANCELLED else False

    @property
    def is_done(self):
        with self._mutex:
            return self._is_done

    @property
    def result(self):
        with self._mutex:
            return self._result

    @property
    def duration(self):
        with self._mutex:
            self._update_pending_duration()
            self._update_task_duration()
            return self._pending_duration, self._task_duration

    @property
    def exception(self):
        with self._mutex:
            return self._exception

    @property
    def callbacks(self):
        with self._mutex:
            return tuple(self._callbacks)

    @property
    def cancel_flag(self):
        with self._mutex:
            return self._cancel_flag

    @property
    def status(self):
        with self._mutex:
            return self._status

    def collect(self, timeout=None):
        """
        Collect the result of the task.
        Beware, a remote exception as well as CancelledError,
        or TimeoutError exceptions may be raised.

        [param]
        - timeout: None or a timeout value (int or float) in seconds.

        [return]
        Returns the collected result

        [except]
        - TimeoutError: raised when timeout expires
        - CancelledError: raised when a task got cancelled
        - Exception: any remote exception. Note that the `__context__`
        attribute of a remote exception contains an instance of
        the RemoteError class. The RemoteError class exposes via
        its exc_chain, the exception chain. Applying the builtin
        'str' function on the RemoteError object will return
        the remote traceback as a string.
        """
        r = self._on_done_event.wait(timeout=timeout)
        if r is False:
            raise TimeoutError
        with self._mutex:
            if self._exception is not None:
                raise self._exception
            if self._status == Status.CANCELLED:
                raise errors.CancelledError
            return self._result

    def wait(self, timeout=None):
        """
        Wait (blocking) for the future to be done (completed, failed, or cancelled).

        [param]
        - timeout: None or a timeout value (int or float) in seconds.

        [return]
        Returns True if the future is done in the provided timeout range.
        """
        return self._on_done_event.wait(timeout=timeout)

    def add_callback(self, callback):
        """
        Add one callback that accepts the future as argument.

        [param]
        - callback: the callback to add
        """
        with self._mutex:
            self._callbacks.append(callback)
            if not self._is_done:
                return
            self._execute_callbacks((callback, ))

    def add_callbacks(self, callbacks):
        """
        Add a sequence of callbacks that accept the future as argument.

        [param]
        - callbacks: sequence of callbacks
        """
        with self._mutex:
            self._callbacks.extend(callbacks)
            if not self._is_done:
                return
            self._execute_callbacks(callbacks)

    def remove_callback(self, callback):
        """
        Remove a callback

        [param]
        - callback: the callback to remove
        """
        with self._mutex:
            self._callbacks = [item for item in self._callbacks if item != callback]

    def remove_callbacks(self, callbacks):
        """
        Remove a sequence of callbacks

        [param]
        - callbacks: sequence of callbacks
        """
        with self._mutex:
            self._callbacks = [item for item in self._callbacks if item not in callbacks]

    def cancel(self):
        """
        Tries to cancel the linked task.
        Note that right after, the cancel_flag property is set to True
        and later, the cancelled property will be set to True or not.
        """
        with self._mutex:
            self._cancel_flag = True

    def set_status(self, status, instant=None):
        """Private method ! Don't call it !"""
        instant = time.monotonic() if instant is None else instant
        with self._mutex:
            if self._is_done:
                raise errors.InvalidStateError
            self._status = status
            self._status_to_time[status] = instant
            if status in (Status.COMPLETED, Status.FAILED,
                          Status.CANCELLED):
                self._is_done = True
                self._on_done_event.set()
                self._execute_callbacks(self._callbacks)

    def set_result(self, result, instant=None):
        """Private method ! Don't call it !"""
        instant = time.monotonic() if instant is None else instant
        with self._mutex:
            if self._is_done:
                raise errors.InvalidStateError
            self._result = result
            self.set_status(Status.COMPLETED, instant)

    def set_exception(self, exc, instant=None):
        """Private method ! Don't call it !"""
        instant = time.monotonic() if instant is None else instant
        with self._mutex:
            if self._is_done:
                raise errors.InvalidStateError
            self._exception = exc
            self.set_status(Status.FAILED, instant)

    def _execute_callbacks(self, callbacks):
        for callback in callbacks:
            try:
                callback(self)
            except Exception as e:
                msg = "Exception while calling callback for future {}".format(repr(self))
                misc.LOGGER.exception(msg)

    def _update_pending_duration(self):
        with self._mutex:
            instant_a = self._status_to_time.get(Status.PENDING)
            if not instant_a:
                return
            for status in (Status.COMPLETED, Status.FAILED, Status.CANCELLED):
                instant_b = self._status_to_time.get(status)
                if not instant_b:
                    continue
                self._pending_duration = instant_b - instant_a
                break

    def _update_task_duration(self):
        with self._mutex:
            instant_a = self._status_to_time.get(Status.RUNNING)
            if not instant_a:
                return
            for status in (Status.COMPLETED, Status.FAILED):
                instant_b = self._status_to_time.get(status)
                if not instant_b:
                    continue
                self._task_duration = instant_b - instant_a
                break


@unique
class Status(Enum):
    PENDING = 1
    RUNNING = 2
    COMPLETED = 3
    FAILED = 4
    CANCELLED = 5



"""Misc functions and classes."""
import os
import sys
import time
import itertools
import functools
import logging
from copy import copy
from traceback import format_exception
from queue import Empty as EmptyQueueError
from asyncpal import errors


__all__ = ["split_map_task", "split_starmap_task",
           "get_chunks", "get_remote_traceback",
           "Countdown", "LOGGER"]


LOGGER = logging.getLogger("asyncpal")


def ensure_args_kwargs(args=None, kwargs=None):
    args = args if args else tuple()
    kwargs = kwargs if kwargs else dict()
    return args, kwargs


def get_chunks(iterable, chunk_size):
    """
    Split an iterable into chunks

    [param]
    - iterable: the iterable to split
    - chunk_size: max length for a chunk

    [return]
    Returns an iterator
    """
    iterator = iter(iterable)
    return iter(lambda: tuple(itertools.islice(iterator, chunk_size)),
                tuple())


def drain_queue(q):
    """Private function !"""
    while True:
        try:
            q.get_nowait()
        except EmptyQueueError as e:
            break


def iterate_queue(q):
    while True:
        try:
            yield q.get_nowait()
        except EmptyQueueError as e:
            break


def get_cpu_count():
    if sys.version_info >= (3, 13):
        return os.process_cpu_count()
    else:
        return os.cpu_count()


def add(a, b):
    return a + b


class Countdown:
    """Class to continually update a timeout as time passes.
    Used a lot by pools for map operations"""
    def __init__(self, timeout):
        """Init.

        [param]
        - timeout: None or a timeout (int or float) value in seconds"""
        self._timeout = timeout
        self._instant_a = time.monotonic()
        if timeout:
            self._instant_b = self._instant_a + timeout
        else:
            self._instant_b = None

    @property
    def timeout(self):
        return self._timeout

    @property
    def instant_a(self):
        """Instant alpha: the instant the Countdown object is created"""
        return self._instant_a

    @property
    def instant_b(self):
        """Instant beta: the exact instant the timeout expires.
        Is None if the original timeout is None"""
        return self._instant_b

    @staticmethod
    def get_instant(self):
        """Returns an instant value"""
        return time.monotonic()

    def check(self):
        """Returns a new timeout value at each call"""
        if self._instant_b is None:
            return self._timeout
        x = self._instant_b - time.monotonic()
        return x if x > 0 else 0


def split_map_task(target, *iterables, chunk_size=1):
    """
    Split a map task into subtasks that don't take any arguments.
    The iterables are chunked according to chunk_size.

    [param]
    - target: the callable task
    - iterables: the map iterables to pass to target
    - chunk_size: the max length of a chunk

    [yield]
    Iteratively get a subtask that don't accept any arguments
    """
    for chunk in get_chunks(zip(*iterables), chunk_size):
        yield functools.partial(_subtask, target, chunk)


def split_starmap_task(target, iterable, chunk_size=1):
    """Split a starmap task into subtasks that don't take any arguments.
    The iterable is chunked according to chunk_size.

    [param]
    - target: the callable task
    - iterable: a sequence of tuples each representing args to pass to target
    - chunk_size: the max length of a chunk

    [yield]
    Iteratively get a subtask that don't accept any arguments
    """
    for chunk in get_chunks(iterable, chunk_size):
        yield functools.partial(_subtask, target, chunk)


def _subtask(target, chunk):
    return [target(*args) for args in chunk]


def get_remote_traceback(exc):
    """Return the remote traceback string"""
    while True:
        cause, context = exc.__cause__, exc.__context__
        if cause is None and context is None:
            return
        if isinstance(context, errors.RemoteTraceback):
            traceback_str = "".join(context.traceback_lines)
            return traceback_str.strip()
        if cause is not None:
            exc = cause
        elif context is not None:
            exc = context
        else:
            return


class RemoteExceptionWrapper:
    def __init__(self, exc):
        self._exc = exc
        # get exc chain
        self._exc_chain = get_exc_chain(exc)
        # get traceback string
        self._traceback_lines = format_exception(type(exc), exc, exc.__traceback__)

    @property
    def exc_chain(self):
        return self._exc_chain

    def unwrap(self):
        last_exc, _ = self._exc_chain[-1]
        last_exc.__context__ = errors.RemoteTraceback(self._traceback_lines)
        exc = None
        last_exc = None
        has_cause = None
        for x, y in self._exc_chain:
            if exc is None:
                exc = last_exc = x
                has_cause = y
                continue
            if has_cause:
                last_exc.__cause__ = x
            else:
                last_exc.__context__ = x
            last_exc = x
            has_cause = y
        return exc


def get_exc_chain(exc):
    exc_chain = list()
    while exc is not None:
        cause = exc.__cause__
        context = None if exc.__suppress_context__ else exc.__context__
        has_cause = False if cause is None else True
        exc_info = (copy(exc), has_cause)
        exc_chain.append(exc_info)
        if cause is not None:
            exc = cause
        elif context is not None:
            exc = context
        else:
            exc = None
    return exc_chain



"""All public Classes, Functions, and Constants"""
from asyncpal.pool import IDLE_TIMEOUT
from asyncpal.misc import (LOGGER, get_chunks, split_map_task,
                           split_starmap_task, get_remote_traceback,
                           Countdown)
from asyncpal.future import Future, FutureFilter, Status, as_done, wait, collect
from asyncpal.pool.threadpool import (Pool, ThreadPool, SingleThreadPool,
                                      DualThreadPool, TripleThreadPool,
                                      QuadThreadPool)
from asyncpal.pool.processpool import (ProcessPool, SingleProcessPool,
                                       DualProcessPool, TripleProcessPool,
                                       QuadProcessPool, MP_CONTEXT,
                                       WINDOWS_MAX_PROCESS_WORKERS)
from asyncpal.errors import (Error, BrokenPoolError,
                             InitializerError, FinalizerError,
                             InvalidStateError, CancelledError)


__all__ = ["Pool", "ThreadPool", "ProcessPool",
           "SingleThreadPool", "SingleProcessPool",
           "DualThreadPool", "DualProcessPool",
           "TripleThreadPool", "TripleProcessPool",
           "QuadThreadPool", "QuadProcessPool",
           "Future", "FutureFilter", "Status", "Countdown",
           "as_done", "wait", "collect", "split_map_task",
           "split_starmap_task", "get_chunks",
           "get_remote_traceback",
           "IDLE_TIMEOUT", "MP_CONTEXT", "LOGGER",
           "WINDOWS_MAX_PROCESS_WORKERS",
           "Error", "BrokenPoolError",
           "InitializerError", "FinalizerError",
           "InvalidStateError", "CancelledError"]



import time
from enum import Enum, unique
from queue import Empty as EmptyQueueError
from asyncpal import errors, misc
from asyncpal.worker import Worker, WorkerType

__all__ = []

@unique
class MessageTag(Enum):
    RUNNING = 1  # tag, task_id
    RESULT = 2  # tag, task_id, result, duration
    EXCEPTION = 3  # tag, task_id, exception
    WORKER_EXCEPTION = 4  # tag, worker_id, exception
    SHUTDOWN = 5  # tag, worker_id


class ProcessWorker(Worker):
    def __init__(self, worker_id, worker_name, task_queue, message_queue,
                 idle_timeout, initializer, init_args, init_kwargs, finalizer,
                 final_args, final_kwargs, max_tasks_per_worker,
                 mp_context):
        super().__init__(WorkerType.PROCESS, worker_id, worker_name,
                         task_queue, idle_timeout, initializer,
                         init_args, init_kwargs, finalizer, final_args,
                         final_kwargs, max_tasks_per_worker)
        self._message_queue = message_queue
        self._mp_context = mp_context
        self._process = None
        self._mutex = self._mp_context.RLock()
        self._is_busy_event = self._mp_context.Event()

    @property
    def message_queue(self):
        return self._message_queue

    @property
    def process(self):
        return self._process

    def run(self):
        with self._mutex:
            if self._process is not None:
                return False
            args = (self._worker_id, self._worker_name, self._task_queue,
                    self._message_queue, self._idle_timeout, self._initializer,
                    self._init_args, self._init_kwargs, self._finalizer,
                    self._final_args, self._final_kwargs,
                    self._max_tasks_per_worker, self._is_busy_event)
            self._process = self._mp_context.Process(name=self._worker_name,
                                                     target=runner, args=args,
                                                     daemon=False)
            self._process.start()
            return True

    def is_alive(self):
        with self._mutex:
            if self._process is None:
                return False
            return self._process.is_alive()

    def is_busy(self):
        return self._is_busy_event.is_set()

    def join(self, timeout=None):
        with self._mutex:
            if self._process is None:
                return False
            self._process.join(timeout=timeout)
            return not self._process.is_alive()


def runner(worker_id, worker_name, task_queue, message_queue, idle_timeout,
           initializer, init_args, init_kwargs,
           finalizer, final_args, final_kwargs,
           max_tasks_per_worker, is_busy_event):
    try:
        if initializer is not None:
            run_initializer(worker_name, initializer, *init_args, **init_kwargs)
        loop(task_queue, message_queue, idle_timeout,
             max_tasks_per_worker, is_busy_event)
        if finalizer is not None:
            run_finalizer(worker_name, finalizer, *final_args, **final_kwargs)
    except BaseException as e:
        misc.LOGGER.critical("Exception in worker", exc_info=True)
        is_busy_event.clear()
        exc = misc.RemoteExceptionWrapper(e)
        e.__traceback__ = e.__cause__ = e.__context__ = None
        msg = (MessageTag.WORKER_EXCEPTION, worker_id, exc)  # WORKER ERROR
        message_queue.put(msg)
    else:
        is_busy_event.clear()
        msg = (MessageTag.SHUTDOWN, worker_id)  # SHUTDOWN
        message_queue.put(msg)


def loop(task_queue, message_queue, idle_timeout,
         max_tasks_per_worker, is_busy_event):
    task_count = 0
    while True:
        if max_tasks_per_worker and max_tasks_per_worker == task_count:
            break
        try:
            task = task_queue.get(block=True, timeout=idle_timeout)
        except EmptyQueueError as e:
            break
        if task is None:
            break
        is_busy_event.set()
        run_task(task, message_queue)
        is_busy_event.clear()
        # task_queue.get might block too long, not giving time to free resource
        # pointed to by 'task', therefore let's free resource as soon as possible
        # by deleting 'task'
        del task
        task_count += 1


def run_task(task, message_queue):
    task_id, target, args, kwargs = task
    # SET RUNNING STATUS
    msg = (MessageTag.RUNNING, task_id, time.monotonic())
    message_queue.put(msg)
    try:
        result = target(*args, **kwargs)
    except BaseException as e:
        # SET EXCEPTION
        exc = misc.RemoteExceptionWrapper(e)
        e.__traceback__ = e.__cause__ = e.__context__ = None
        msg = (MessageTag.EXCEPTION, task_id, exc,
               time.monotonic())
        message_queue.put(msg)
    else:
        # SET RESULT
        msg = (MessageTag.RESULT, task_id, result,
               time.monotonic())
        message_queue.put(msg)


def run_initializer(worker_name, initializer, *init_args, **init_kwargs):
    try:
        initializer(*init_args, **init_kwargs)
    except BaseException as e:
        msg = "Initialization error in '{}'.".format(worker_name)
        raise errors.InitializerError(msg) from e


def run_finalizer(worker_name, finalizer, *final_args, **final_kwargs):
    try:
        finalizer(*final_args, **final_kwargs)
    except BaseException as e:
        msg = "Finalization error in '{}'.".format(worker_name)
        raise errors.FinalizerError(msg) from e


import unittest


def run_tests():
    test_loader = unittest.defaultTestLoader
    test_suite = test_loader.discover("tests", pattern="test_*.py")
    test_runner = unittest.TextTestRunner(verbosity=0)
    test_runner.run(test_suite)


if __name__ == "__main__":
    run_tests()



import time
import threading
from queue import Empty as EmptyQueueError
from asyncpal import errors, misc
from asyncpal.future import Status
from asyncpal.worker import Worker, WorkerType

__all__ = []

class ThreadWorker(Worker):
    def __init__(self, worker_id, worker_name, task_queue, idle_timeout, initializer,
                 init_args, init_kwargs, finalizer, final_args, final_kwargs,
                 max_tasks_per_worker, on_shutdown, on_exception):
        super().__init__(WorkerType.THREAD, worker_id, worker_name,
                         task_queue, idle_timeout, initializer,
                         init_args, init_kwargs, finalizer, final_args, final_kwargs,
                         max_tasks_per_worker)
        self._on_shutdown = on_shutdown
        self._on_exception = on_exception
        self._thread = None
        self._mutex = threading.RLock()
        self._is_busy_event = threading.Event()

    @property
    def thread(self):
        return self._thread

    @property
    def on_shutdown(self):
        return self._on_shutdown

    @property
    def on_exception(self):
        return self._on_exception

    def run(self):
        with self._mutex:
            if self._thread is not None:
                return False
            args = (self._worker_id, self._worker_name, self._task_queue, self._idle_timeout,
                    self._initializer, self._init_args, self._init_kwargs,
                    self._finalizer, self._final_args, self._final_kwargs,
                    self._max_tasks_per_worker, self._is_busy_event,
                    self._on_shutdown, self._on_exception)
            self._thread = threading.Thread(name=self._worker_name, target=runner,
                                            args=args, daemon=False)
            self._thread.start()
            return True

    def is_alive(self):
        with self._mutex:
            if self._thread is None:
                return False
            return self._thread.is_alive()

    def is_busy(self):
        return self._is_busy_event.is_set()

    def join(self, timeout=None):
        with self._mutex:
            if self._thread is None:
                return False
            self._thread.join(timeout=timeout)
            return not self._thread.is_alive()


def runner(worker_id, worker_name, task_queue, idle_timeout,
           initializer, init_args, init_kwargs,
           finalizer, final_args, final_kwargs,
           max_tasks_per_worker, is_busy_event,
           on_shutdown, on_exception):
    try:
        if initializer is not None:
            run_initializer(worker_name, initializer, *init_args, **init_kwargs)
        loop(task_queue, idle_timeout, max_tasks_per_worker, is_busy_event)
        if finalizer is not None:
            run_finalizer(worker_name, finalizer, *final_args, **final_kwargs)
    except BaseException as e:
        misc.LOGGER.critical("Exception in worker", exc_info=True)
        is_busy_event.clear()
        on_exception(worker_id, e)
    else:
        is_busy_event.clear()
        on_shutdown(worker_id)


def loop(task_queue, idle_timeout, max_tasks_per_worker, is_busy_event):
    task_count = 0
    while True:
        if max_tasks_per_worker and max_tasks_per_worker == task_count:
            break
        try:
            task = task_queue.get(block=True, timeout=idle_timeout)
        except EmptyQueueError as e:
            break
        if task is None:
            break
        is_busy_event.set()
        run_task(task)
        is_busy_event.clear()
        # task_queue.get might block too long, not giving time to free resource
        # pointed to by 'task', therefore let's free resource as soon as possible
        # by deleting 'task'
        del task
        task_count += 1


def run_task(task):
    future, target, args, kwargs = task
    # SET RUNNING STATUS
    future.set_status(Status.RUNNING, time.monotonic())
    if future.cancel_flag:
        future.set_status(Status.CANCELLED)
        return
    try:
        result = target(*args, **kwargs)
    except BaseException as e:
        # SET EXCEPTION
        future.set_exception(e, time.monotonic())
    else:
        # SET RESULT
        future.set_result(result, time.monotonic())


def run_initializer(worker_name, initializer, *init_args, **init_kwargs):
    try:
        initializer(*init_args, **init_kwargs)
    except BaseException as e:
        msg = "Initialization error in '{}'.".format(worker_name)
        raise errors.InitializerError(msg) from e


def run_finalizer(worker_name, finalizer, *final_args, **final_kwargs):
    try:
        finalizer(*final_args, **final_kwargs)
    except BaseException as e:
        msg = "Finalization error in '{}'.".format(worker_name)
        raise errors.FinalizerError(msg) from e



"""ProcessPool, SingleProcessPool, DualProcessPool,
TripleProcessPool, and QuadProcessPool are defined here"""
import sys
import threading
import queue
from asyncpal import errors
from asyncpal import misc
from asyncpal.future import Status
from asyncpal.worker.processworker import ProcessWorker, MessageTag
from asyncpal.pool import Pool, WorkerType, IDLE_TIMEOUT, MP_CONTEXT, WINDOWS_MAX_PROCESS_WORKERS


__all__ = ["ProcessPool", "SingleProcessPool", "DualProcessPool",
           "TripleProcessPool", "QuadProcessPool"]


class ProcessPool(Pool):
    """The ProcessPool class for parallelism."""
    def __init__(self, max_workers=None, *, name="ProcessPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None, mp_context=None):
        """
        Initialization.

        [param]
        - max_workers: the maximum number of workers. Defaults to CPU count. On Windows,
            the maximum number that is accepted for the max_workers is 60.
        - name: the name of the pool. Defaults to the class name
        - idle_timeout: None or a timeout value in seconds.
            The idle timeout tells how much time an inactive worker
            can sleep before it closes. This helps the pool to shrink
            when there isn't much of tasks.
            If you set None, the pool will never shrink. each worker before it closes
        - initializer: a function that will get called at the start of each worker
        - init_args: arguments (list) to pass to the initializer
        - init_kwargs: keyword arguments (dict) to pass to the initializer
        - finalizer: a function that will get called when the worker is going to close
        - final_args: arguments (list) to pass to the finalizer
        - final_kwargs: keyword arguments (dict) to pass to the finalizer
        - max_tasks_per_worker: Maximum number of tasks a worker is allowed to do
            before it closes.
        - mp_context: the multiprocessing context.
            Defaults to multiprocessing.get_context("spawn")
        """
        if max_workers:
            if sys.platform == "win32" and max_workers > WINDOWS_MAX_PROCESS_WORKERS:
                msg = "Max process workers on Windows is {}".format(WINDOWS_MAX_PROCESS_WORKERS)
                raise ValueError(msg)
        else:
            x = misc.get_cpu_count()
            max_workers = x if x else 1
        worker_type = WorkerType.PROCESS
        mp_context = MP_CONTEXT if mp_context is None else mp_context
        task_queue = queue.SimpleQueue()
        mp_task_queue = mp_context.Queue(maxsize=max_workers+1)
        message_queue = mp_context.SimpleQueue()
        super().__init__(worker_type, max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer, init_args=init_args,
                         init_kwargs=init_kwargs, finalizer=finalizer,
                         final_args=final_args, final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         task_queue=task_queue, mp_task_queue=mp_task_queue,
                         message_queue=message_queue, mp_context=mp_context)
        self._spawn_filter_thread()
        self._spawn_message_thread()

    def _create_worker(self):
        worker_id = self._monotonic_worker_count
        worker_name = "asyncpal-{}-process-worker-{}".format(self._name, worker_id)
        with self._vars_lock:
            worker = ProcessWorker(worker_id, worker_name, self._mp_task_queue,
                                   self._message_queue,
                                   self._idle_timeout, self._initializer,
                                   self._init_args, self._init_kwargs,
                                   self._finalizer, self._final_args,
                                   self._final_kwargs,
                                   self._max_tasks_per_worker,
                                   self._mp_context)
            return worker

    def _spawn_message_thread(self):
        thread_name = "asyncpal-{}-MessageThread".format(self._name)
        self._message_thread = threading.Thread(name=thread_name,
                                                target=self._consume_message_queue,
                                                daemon=False)
        self._message_thread.start()

    def _consume_message_queue(self):
        # special loop working inside a thread to support
        # Process Workers by consuming/dispatching the messages
        # they put in the message_queue
        while True:
            message = self._message_queue.get()
            if message is None:
                break
            try:
                tag = message[0]
                if tag in (MessageTag.RUNNING, MessageTag.RESULT,
                           MessageTag.EXCEPTION):
                    self._update_future(message)
                elif tag == MessageTag.SHUTDOWN:
                    self._on_worker_shutdown(message[1])
                elif tag == MessageTag.WORKER_EXCEPTION:
                    worker_id, exc_wrapper = message[1:]
                    exc = exc_wrapper.unwrap()
                    self._on_worker_exception(worker_id, exc)
                else:
                    msg = "Invalid message tag: {}".format(tag)
                    e = errors.Error(msg)
                    raise e
            except BaseException as e:
                with self._vars_lock:
                    self._stored_exception = e
                raise e
            # message_queue.get might block too long, not giving time to free resource
            # pointed to by 'message', therefore let's free resource as soon as possible
            # by deleting 'message'
            del message

    def _update_future(self, message):
        # this method is solely called by the message_thread
        # that works only for Process Workers to dispatch results/notifs
        tag, task_id = message[0:2]
        with self._futures_lock:
            future = self._stored_futures[task_id]
        if tag == MessageTag.RUNNING:
            instant = message[2]
            future.set_status(Status.RUNNING, instant)
        elif tag == MessageTag.RESULT:
            result, instant = message[2:]
            future.set_result(result, instant)
        elif tag == MessageTag.EXCEPTION:
            exc_wrapper, instant = message[2:]
            exc = exc_wrapper.unwrap()
            future.set_exception(exc, instant)
        if tag in (MessageTag.RESULT, MessageTag.EXCEPTION):
            with self._futures_lock:
                del self._stored_futures[task_id]

    def __reduce__(self):
        msg = "A pool object cannot be pickled"
        raise NotImplementedError(msg)


class SingleProcessPool(ProcessPool):
    """Fixed-size process pool. This pool can spawn only 1 worker"""
    def __init__(self, *, name="SingleProcessPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None, mp_context=None):
        max_workers = 1
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer, init_args=init_args,
                         init_kwargs=init_kwargs, finalizer=finalizer,
                         final_args=final_args, final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         mp_context=mp_context)


class DualProcessPool(ProcessPool):
    """Fixed-size process pool. This pool can spawn up to 2 workers"""
    def __init__(self, *, name="DualProcessPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None, mp_context=None):
        max_workers = 2
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer, init_args=init_args,
                         init_kwargs=init_kwargs, finalizer=finalizer,
                         final_args=final_args, final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         mp_context=mp_context)


class TripleProcessPool(ProcessPool):
    """Fixed-size process pool. This pool can spawn up to 3 workers"""
    def __init__(self, *, name="TripleProcessPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None, mp_context=None):
        max_workers = 3
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer, init_args=init_args,
                         init_kwargs=init_kwargs, finalizer=finalizer,
                         final_args=final_args, final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         mp_context=mp_context)


class QuadProcessPool(ProcessPool):
    """Fixed-size process pool. This pool can spawn up to 4 workers"""
    def __init__(self, *, name="QuadProcessPool",
                 idle_timeout=IDLE_TIMEOUT, initializer=None, init_args=None,
                 init_kwargs=None, finalizer=None, final_args=None,
                 final_kwargs=None, max_tasks_per_worker=None, mp_context=None):
        max_workers = 4
        super().__init__(max_workers=max_workers, name=name,
                         idle_timeout=idle_timeout,
                         initializer=initializer, init_args=init_args,
                         init_kwargs=init_kwargs, finalizer=finalizer,
                         final_args=final_args, final_kwargs=final_kwargs,
                         max_tasks_per_worker=max_tasks_per_worker,
                         mp_context=mp_context)



import time


def divide(a, b, sleep=0):
    if sleep:
        time.sleep(sleep)
    return a/b


def add(a, b, sleep=0):
    if sleep:
        time.sleep(sleep)
    return a + b


def square(x, sleep=0):
    if sleep:
        time.sleep(sleep)
    return x**2


def get_worker_exception(pool):
    try:
        pool.check()
    except Exception as e:
        return e
